# -*- coding: utf-8 -*-
# Copyright (C) Katelyn Breivik (2017 - 2019)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`Match`
"""

import numpy as np
import pandas as pd
import astropy.stats as astroStats
import warnings
from cosmic.utils import dat_transform


__author__ = 'Katelyn Breivik <katie.breivik@gmail.com>'
__credits__ = 'Scott Coughlin <scott.coughlin@ligo.org>'
__all__ = ['match', 'perform_convergence']


def match(dataCm):
    """Performs the Match calculation in Eq. 1 of Breivik & Larson (2018)

        Parameters
        ----------
        dataCm : list
            List of two cumulative data sets for a single paramter

        Returns
        -------
        match : list
            List of matches for each cumulative data set

        binwidth : float
            Binwidth of histograms used for match computation
    """

    # DEFINE A LIST TO HOLD THE BINNED DATA:
    histo = [[], []]
    histoBinEdges = [[], []]
    # COMPUTE THE BINWIDTH FOR THE MOST COMPLETE DATA SET:
    # NOTE: THIS WILL BE THE BINWIDTH FOR ALL THE HISTOGRAMS IN THE HISTO LIST
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="divide by zero encountered in double_scalars")
        try:
            bw, binEdges = astroStats.knuth_bin_width(np.array(dataCm[0]), return_bins=True)
        except:
            bw, binEdges = astroStats.scott_bin_width(np.array(dataCm[0]), return_bins=True)
    if bw < 1e-4:
        bw = 1e-4
        binEdges = np.arange(binEdges[0], binEdges[-1], bw)

    # BIN THE DATA:
    for i in range(2):
        histo[i], histoBinEdges[i] = astroStats.histogram(dataCm[i], bins = binEdges, density = True)
    # COMPUTE THE MATCH:
    nominator = []
    denominator1 = []
    denominator2 = []
    nominatorSum = []
    denominator1Sum = []
    denominator2Sum = []

    histo2 = histo[1]
    histo1 = histo[0]

    for j in range(len(histo1)):
        nominator.append(histo1[j]*histo2[j])
        denominator1.append((histo1[j]*histo1[j]))
        denominator2.append((histo2[j]*histo2[j]))
    nominatorSum.append(np.sum(nominator))
    denominator1Sum.append(np.sum(denominator1))
    denominator2Sum.append(np.sum(denominator2))

    nominatorSum = np.array(nominatorSum)
    denominator1Sum = np.array(denominator1Sum)
    denominator2Sum = np.array(denominator2Sum)

    binwidth = binEdges[1]-binEdges[0]
    if binwidth < 1e-7:
        match = 1e-9
    else:
        match = np.log10(1-nominatorSum/np.sqrt(denominator1Sum*denominator2Sum))


    return match[0], binwidth;

def perform_convergence(conv_params, conv_1, conv_2, log_file):
    """Performs the convergence calculations for each convergence parameter
       and binary state

       Parameters
       ----------
       conv_params : dict
           List of user supplied convergence parameters
       conv_1 : DataFrame
           Cumulative data set of conv arrays
       conv_2 : DataFrame
           Most recent data set of conv array from most recent BSE run
       log_file : file write
           File to log matches or if the convergence params are not appropriate
           e.g. eccentricity for a disrupted system

       Returns
       -------
       match : array
           Matches for each cumulative data set
    """
    match_all = []
    for conv_param in conv_params:
        if (conv_param == 'ecc') and (np.all(conv_1[conv_param] < 1e-7)) and (np.all(conv_1[conv_param] >=0.0)) and (bin_state == 0):
            log_file.write('{0} is circular for all conv binaries\n'.format(conv_param))
            match_all.append(-9)
        elif (conv_param == 'ecc') and (np.all(conv_1[conv_param] == -1.0)):
            log_file.write('{0} is the same for all disrupted binaries\n'.format(conv_param))
            match_all.append(-9)
        elif (conv_param == 'ecc'):
            conv_1_ecc = conv_1.loc[conv_1.ecc > 0]
            conv_2_ecc = conv_2.loc[conv_2.ecc > 0]
            if len(conv_2_ecc) == len(conv_1_ecc):
               conv_2_ecc = conv_2[:int(len(conv_2_ecc)/2)]
            match_compute, bw = match([dat_transform(conv_1_ecc, [conv_param])[0].tolist(),\
                                       dat_transform(conv_2_ecc, [conv_param])[0].tolist()])
            match_all.append(match_compute)

        elif (conv_param == 'porb') and ((np.all(conv_1[conv_param] == 0.0)) or (np.all(conv_1[conv_param] == -1.0))):
            log_file.write('{0} is the same for all converging binaries\n'.format(conv_param))
            match_all.append(-9)
        elif (conv_param == 'sep') and ((np.all(conv_1[conv_param] == 0.0)) or (np.all(conv_1[conv_param] == -1.0))):
            log_file.write('{0} is the same for all converging binaries\n'.format(conv_param))
            match_all.append(-9)
        else:
            if len(conv_2) == len(conv_1):
               conv_2 = conv_2[:int(len(conv_2)/2)]
            match_compute, bw = match([dat_transform(conv_1, [conv_param])[0].tolist(),\
                                       dat_transform(conv_2, [conv_param])[0].tolist()])
            match_all.append(match_compute)

    log_file.write('matches for converging population are are: {0}\n'.format(match_all))
    log_file.write('Number of binaries in converging population is: {0}\n'.format(len(conv_1)))
    log_file.write('Binwidth is: {0}\n'.format(bw))

    return match_all



# -*- coding: utf-8 -*-
# Copyright (C) Scott Coughlin (2017 - 2019)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`InitialBinaryTable`
"""

import numpy as np
import multiprocessing as mp
import math
import random
import scipy.integrate
import sys

from astropy.table import Table, Column
from astropy import units
import pandas as pd

from cosmic.utils import idl_tabulate, rndm

__author__ = 'Katelyn Breivik <katie.breivik@gmail.com>'
__credits__ = 'Scott Coughlin <scott.coughlin@ligo.org>'
__all__ = ['InitialBinaryTable']


G = 6.67384*math.pow(10, -11.0)
c = 2.99792458*math.pow(10, 8.0)
parsec = 3.08567758*math.pow(10, 16)
Rsun = 6.955*math.pow(10, 8)
Msun = 1.9891*math.pow(10,30)
day = 86400.0
rsun_in_au = 215.0954
day_in_year = 365.242
sec_in_day = 86400.0
sec_in_hour = 3600.0
hrs_in_day = 24.0
sec_in_year = 3.15569*10**7.0
Tobs = 3.15569*10**7.0
geo_mass = G/c**2

INITIAL_CONDITIONS_COLUMNS = ['kstar_1', 'kstar_2', 'mass1_binary', 'mass2_binary', 'porb', 'ecc',
                             'metallicity', 'tphysf', 'mass0_1', 'mass0_2',
                             'rad1', 'rad2', 'lumin1', 'lumin2', 'massc1', 'massc2',
                             'radc1', 'radc2', 'menv1', 'menv2', 'renv1', 'renv2',
                             'ospin1', 'ospin2', 'b_0_1', 'b_0_2', 'bacc1', 'bacc2',
                             'tacc1', 'tacc2', 'epoch1', 'epoch2', 'tms1', 'tms2',
                             'bhspin1','bhspin2', 'tphys']

INITIAL_CONDITIONS_MISC_COLUMNS = ['binfrac']

if sys.version_info.major == 2 and sys.version_info.minor == 7:
    INITIAL_CONDITIONS_COLUMNS_ALL = INITIAL_CONDITIONS_COLUMNS[:]
else:
    INITIAL_CONDITIONS_COLUMNS_ALL = INITIAL_CONDITIONS_COLUMNS.copy()

INITIAL_CONDITIONS_COLUMNS_ALL.extend(INITIAL_CONDITIONS_MISC_COLUMNS)


class InitialBinaryTable():
    @classmethod
    def InitialBinaries(cls, m1, m2, porb, ecc, tphysf, kstar1, kstar2, metallicity, **kwargs):
        """Create single binary

        Parameters
        ----------
        m1 : float
            Primary mass [Msun]
        m2 : float
            Secondary mass [Msun]
        porb : float
            Orbital period [days]
        ecc : float
            Eccentricity
        tphysf : float
            Time to evolve the binary [Myr]
        kstar1 : array
            0-14 Initial stellar type of the larger object;
            main sequence stars are 0 if m < 0.7 Msun and 1 otherwise
        kstar2 : array
            0-14 Initial stellar type of the smaller object;
            main sequence stars are 0 if m < 0.7 Msun and 1 otherwise
        metallicity : float
            Metallicity of the binaries; Z_sun = 0.02

        **kwargs

            binfrac : float
                System-specific probability of the primary star being in a binary

            mass0_1,mass0_2,rad1,rad2,lumin1,lumin2,
            massc1,massc2,radc1,radc2,menv1,menv2,renv1,renv2,
            ospin1,ospin2,b_0_1,b_0_2,bacc1,bacc2,
            tacc1,tacc2,epoch1,epoch2,tms1,tms2
            bhspin1,bhspin2

        Returns
        -------
        InitialBinaries : DataFrame
            Single binary initial conditions

        """
        # System-specific probability of the primary star being in a binary
        binfrac = kwargs.pop('binfrac', np.ones(np.array(m1).size))

        # The following are in general not needed for COSMIC, however
        # if you wish to start a binary at a very specific point in its evolution
        # these kwarg arguments can help you do this.
        # For instance the Globular Cluster code CMC requires this behavior.
        mass0_1 = kwargs.pop('mass0_1', m1)
        mass0_2 = kwargs.pop('mass0_2', m2)
        rad1 = kwargs.pop('rad1', np.zeros(np.array(m1).size))
        rad2 = kwargs.pop('rad2', np.zeros(np.array(m1).size))
        lumin1 = kwargs.pop('lumin1', np.zeros(np.array(m1).size))
        lumin2 = kwargs.pop('lumin2', np.zeros(np.array(m1).size))
        massc1 = kwargs.pop('massc1', np.zeros(np.array(m1).size))
        massc2 = kwargs.pop('massc2', np.zeros(np.array(m1).size))
        radc1 = kwargs.pop('radc1', np.zeros(np.array(m1).size))
        radc2 = kwargs.pop('radc2', np.zeros(np.array(m1).size))
        menv1 = kwargs.pop('menv1', np.zeros(np.array(m1).size))
        menv2 = kwargs.pop('menv2', np.zeros(np.array(m1).size))
        renv1 = kwargs.pop('renv1', np.zeros(np.array(m1).size))
        renv2 = kwargs.pop('renv2', np.zeros(np.array(m1).size))
        ospin1 = kwargs.pop('ospin1', np.zeros(np.array(m1).size))
        ospin2 = kwargs.pop('ospin2', np.zeros(np.array(m1).size))
        b_0_1 = kwargs.pop('b_0_1', np.zeros(np.array(m1).size))
        b_0_2 = kwargs.pop('b_0_2', np.zeros(np.array(m1).size))
        bacc1 = kwargs.pop('bacc1', np.zeros(np.array(m1).size))
        bacc2 = kwargs.pop('bacc2', np.zeros(np.array(m1).size))
        tacc1 = kwargs.pop('tacc1', np.zeros(np.array(m1).size))
        tacc2 = kwargs.pop('tacc2', np.zeros(np.array(m1).size))
        epoch1 = kwargs.pop('epoch1', np.zeros(np.array(m1).size))
        epoch2 = kwargs.pop('epoch2', np.zeros(np.array(m1).size))
        tms1 = kwargs.pop('tms1', np.zeros(np.array(m1).size))
        tms2 = kwargs.pop('tms2', np.zeros(np.array(m1).size))
        bhspin1 = kwargs.pop('bhspin1', np.zeros(np.array(m1).size))
        bhspin2 = kwargs.pop('bhspin2', np.zeros(np.array(m1).size))
        tphys = kwargs.pop('tphys', np.zeros(np.array(m1).size))

        bin_dat = pd.DataFrame(np.vstack([
                                            kstar1, kstar2,
                                            m1, m2, porb, ecc,
                                            metallicity, tphysf,
                                            mass0_1, mass0_2, rad1, rad2,
                                            lumin1, lumin2, massc1, massc2,
                                            radc1, radc2, menv1, menv2, renv1, renv2,
                                            ospin1, ospin2, b_0_1, b_0_2, bacc1, bacc2,
                                            tacc1, tacc2, epoch1, epoch2, tms1, tms2,
                                            bhspin1,bhspin2,tphys,binfrac,
                                          ]).T,
                               columns = INITIAL_CONDITIONS_COLUMNS_ALL)

        return bin_dat

    @classmethod
    def sampler(cls, format_, *args, **kwargs):
        """Fetch a method to generate an initial binary sample

        Parameters
        ----------
        format : str
            the method name; Choose from 'independent' or 'multidim'

        *args
            the arguments necessary for the registered sample
            method; see help(InitialBinaryTable.sampler('independent')
            to see the arguments necessary for the independent sample
        """
        # standard registered fetch
        from .sampler.sampler import get_sampler
        sampler = get_sampler(format_, cls)
        return sampler(*args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Fetch registration for database queries
"""

import re

from six import string_types

from astropy.io.registry import IORegistryError
from astropy.table import Table

_SAMPLERS = {}

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def register_sampler(data_format, data_class, function, force=False,
                     usage=None):
    """Register a new method to InitialBinaryTable.sampler() for a given format

    Parameters
    ----------
    data_format : `str`
        name of the format to be registered

    data_class : `type`
        the class that the sampler returns

    function : `callable`
        the method to call from :meth:`InitialBinaryTable.sampler`

    force : `bool`, optional
        overwrite existing registration for ``data_format`` if found,
        default: `False`
    """
    key = (data_format, data_class)
    if key not in _SAMPLERS or force:
        _SAMPLERS[key] = (function, usage)
    else:
        raise IORegistryError("Fetcher for format '{0}' and class '{1}' "
                              "has already been " "defined".format(
                                  data_format, data_class))
    _update__doc__(data_class)


def get_sampler(data_format, data_class):
    """Return the :meth:`~InitialBinaryTable.sampler` function for the given format

    Parameters
    ----------
    data_format : `str`
        name of the format

    data_class : `type`
        the class that the sampler returns

    Raises
    ------
    astropy.io.registry.IORegistryError
        if not registration is found matching ``data_format``
    """
    try:
        return _SAMPLERS[(data_format, data_class)][0]
    except KeyError:
        formats = '\n'.join(_SAMPLERS.keys())
        raise IORegistryError("No sampler definer for format %r. "
                              "The available formats are:\n%r"
                              % (data_format, formats))


def _update__doc__(data_class):
    header = "The available named formats are:"
    sampler = data_class.sampler

    # if __doc__ isn't a string, bail-out now
    if not isinstance(sampler.__doc__, string_types):
        return

    # remove the old format list
    lines = sampler.__doc__.splitlines()
    try:
        pos = [i for i, line in enumerate(lines) if header in line][0]
    except IndexError:
        pass
    else:
        lines = lines[:pos]

    # work out the indentation
    matches = [re.search(r'(\S)', line) for line in lines[1:]]
    indent = min(match.start() for match in matches if match)

    # now re-write the format list
    formats = []
    for fmt, cls in sorted(_SAMPLERS, key=lambda x: x[0]):
        if cls is not data_class:
            continue
        usage = _SAMPLERS[(fmt, cls)][1]
        formats.append((
            fmt, '``sampler(%r, %s)``' % (fmt, usage)))
    format_str = Table(rows=formats, names=['Format', 'Basic usage']).pformat(
        max_lines=-1, max_width=80, align=('>', '<'))
    format_str[1] = format_str[1].replace('-', '=')
    format_str.insert(0, format_str[1])
    format_str.append(format_str[0])

    lines.extend([' ' * indent + line for line in [header, ''] + format_str])
    # and overwrite the docstring
    try:
        sampler.__doc__ = '\n'.join(lines)
    except AttributeError:
        sampler.__func__.__doc__ = '\n'.join(lines)



# -*- coding: utf-8 -*-
# Copyright (C) Katelyn Breivik (2017 - 2019)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`independent`
"""

import numpy as np
import multiprocessing as mp
import math
import random
import scipy.integrate

from cosmic.utils import mass_min_max_select

from .sampler import register_sampler
from .. import InitialBinaryTable

from cosmic.utils import idl_tabulate, rndm

__author__ = 'Katelyn Breivik <katie.breivik@gmail.com>'
__credits__ = 'Scott Coughlin <scott.coughlin@ligo.org>'
__all__ = ['get_independent_sampler', 'Sample']


def get_independent_sampler(final_kstar1, final_kstar2, primary_model, ecc_model, SFH_model, binfrac_model, component_age, met, size, **kwargs):
    """Generates an initial binary sample according to user specified models

    Parameters
    ----------
    final_kstar1 : `int or list`
        Int or list of final kstar1

    final_kstar2 : `int or list`
        Int or list of final kstar2

    primary_model : `str`
        Model to sample primary mass; choices include: kroupa93, kroupa01, salpeter55

    ecc_model : `str`
        Model to sample eccentricity; choices include: thermal, uniform

    SFH_model : `str`
        Model to sample star formation history (or birth time); choices include: const, burst, delta_burst

    binfrac_model : `str or float`
        Model for binary fraction; choices include: vanHaaften or a fraction where 1.0 is 100% binaries

    component_age : `float`
        Sets the maximum age of the component; in the case of a delta burst, every binary is evolved for the component age

    met : `float`
        Sets the metallicity of the binary population where solar metallicity is 0.02

    size : `int`
        Size of the population to sample

    Returns
    -------
    InitialBinaryTable : `pandas.DataFrame`
        DataFrame in the format of the InitialBinaryTable

    mass_singles : `float`
        Total mass in single stars needed to generate population

    mass_binaries : `float`
        Total mass in binaries needed to generate population

    n_singles : `int`
        Number of single stars needed to generate a population

    n_binaries : `int`
        Number of binaries needed to generate a population
    """
    if type(final_kstar1) in [int, float]:
        final_kstar1 = [final_kstar1]
    if type(final_kstar2) in [int, float]:
        final_kstar2 = [final_kstar2]
    primary_min, primary_max, secondary_min, secondary_max = mass_min_max_select(final_kstar1, final_kstar2)
    initconditions = Sample()

    #set up multiplier if the mass sampling is inefficient
    multiplier = 1
    mass1_binary = []
    mass2_binary = []
    binfrac = []

    # track the mass in singles and the mass in binaries
    mass_singles = 0.0
    mass_binaries = 0.0

    # track the total number of stars sampled
    n_singles = 0
    n_binaries = 0
    while len(mass1_binary) < size:
        mass1, total_mass1 = initconditions.sample_primary(primary_model, size=size*multiplier)
        mass1_binaries, mass_single, binfrac_binaries = initconditions.binary_select(mass1, binfrac_model=binfrac_model)
        mass2_binaries = initconditions.sample_secondary(mass1_binaries)

        # track the mass sampled
        mass_singles += np.sum(mass_single)
        mass_binaries += np.sum(mass1_binaries)
        mass_binaries += np.sum(mass2_binaries)

        # track the total number sampled
        n_singles += len(mass_single)
        n_binaries += len(mass1_binaries)

        # select out the primaries and secondaries that will produce the final kstars
        ind_select_primary, = np.where((mass1_binaries > primary_min) & (mass1_binaries < primary_max))
        ind_select_secondary, = np.where((mass2_binaries > secondary_min) & (mass2_binaries < secondary_max))
        ind_select = list(set(ind_select_primary).intersection(ind_select_secondary))
        mass1_binary.extend(mass1_binaries[ind_select])
        mass2_binary.extend(mass2_binaries[ind_select])
        binfrac.extend(binfrac_binaries[ind_select])
        # check to see if we should increase the multiplier factor to sample the population more quickly

        if len(mass1_binary) < size/100:
            # well this size clearly is not working time to increase
            # the multiplier by an order of magnitude
            multiplier *= 10

    mass1_binary = np.array(mass1_binary)
    mass2_binary = np.array(mass2_binary)
    binfrac = np.asarray(binfrac)
    ecc =  initconditions.sample_ecc(ecc_model, size = mass1_binary.size)
    porb =  initconditions.sample_porb(mass1_binary, mass2_binary, ecc, size=mass1_binary.size)
    tphysf, metallicity = initconditions.sample_SFH(SFH_model, component_age=component_age, met=met, size = mass1_binary.size)
    metallicity[metallicity < 1e-4] = 1e-4
    metallicity[metallicity > 0.03] = 0.03
    kstar1 = initconditions.set_kstar(mass1_binary)
    kstar2 = initconditions.set_kstar(mass2_binary)

    return InitialBinaryTable.InitialBinaries(mass1_binary, mass2_binary, porb, ecc, tphysf, kstar1, kstar2, metallicity, binfrac=binfrac), mass_singles, mass_binaries, n_singles, n_binaries



register_sampler('independent', InitialBinaryTable, get_independent_sampler,
                 usage="final_kstar1, final_kstar2, binfrac_model, primary_model, ecc_model, SFH_model, component_age, metallicity, size")


class Sample(object):

    # sample primary masses
    def sample_primary(self, primary_model='kroupa93', size=None):
        """Sample the primary mass (always the most massive star) from a user-selected model

        kroupa93 follows Kroupa (1993), normalization comes from
        `Hurley 2002 <https://arxiv.org/abs/astro-ph/0201220>`_
        between 0.08 and 150 Msun
        salpter55 follows
        `Salpeter (1955) <http://adsabs.harvard.edu/abs/1955ApJ...121..161S>`_
        between 0.08 and 150 Msun
        kroupa01 follows Kroupa (2001) <https://arxiv.org/abs/astro-ph/0009005>
        between 0.08 and 100 Msun


        Parameters
        ----------
        primary_model : str, optional
            model for mass distribution; choose from:

            kroupa93 follows Kroupa (1993), normalization comes from
            `Hurley 2002 <https://arxiv.org/abs/astro-ph/0201220>`_
            valid for masses between 0.1 and 100 Msun


            salpter55 follows
            `Salpeter (1955) <http://adsabs.harvard.edu/abs/1955ApJ...121..161S>`_
            valid for masses between 0.1 and 100 Msun

            kroupa01 follows Kroupa (2001), normalization comes from
            `Hurley 2002 <https://arxiv.org/abs/astro-ph/0009005>`_
            valid for masses between 0.1 and 100 Msun

            Default kroupa93
        size : int, optional
            number of initial primary masses to sample
            NOTE: this is set in cosmic-pop call as Nstep

        Returns
        -------
        a_0 : array
            Sampled primary masses
        sampled_mass : float
            Total amount of mass sampled
        """

        if primary_model=='kroupa93':
            total_sampled_mass = 0
            multiplier = 1
            a_0 = np.random.uniform(0.0, 1, size)

            low_cutoff = 0.740074
            high_cutoff = 0.908422

            lowIdx, = np.where(a_0 <= low_cutoff)
            midIdx, = np.where((a_0 > low_cutoff) & (a_0 < high_cutoff))
            highIdx, = np.where(a_0 >= high_cutoff)

            a_0[lowIdx] = rndm(a=0.08, b=0.5, g=-0.3, size=len(lowIdx))
            a_0[midIdx] = rndm(a=0.50, b=1.0, g=-1.2, size=len(midIdx))
            a_0[highIdx] = rndm(a=1.0, b=150.0, g=-1.7, size=len(highIdx))

            total_sampled_mass += np.sum(a_0)

            return a_0, total_sampled_mass

        elif primary_model=='kroupa01':
            total_sampled_mass = 0
            multiplier = 1
            a_0 = np.random.uniform(0.0, 1, size)

            low_cutoff = 0.37148816884988606
            high_cutoff = 0.8496015751162523

            lowIdx, = np.where(a_0 <= low_cutoff)
            midIdx, = np.where((a_0 > low_cutoff) & (a_0 < high_cutoff))
            highIdx, = np.where(a_0 >= high_cutoff)

            a_0[lowIdx] = rndm(a=0.01, b=0.08, g=0.7, size=len(lowIdx))
            a_0[midIdx] = rndm(a=0.08, b=0.5, g=-0.3, size=len(midIdx))
            a_0[highIdx] = rndm(a=0.5, b=100.0, g=-1.3, size=len(highIdx))

            total_sampled_mass += np.sum(a_0)

            return a_0, total_sampled_mass

        elif primary_model=='salpeter55':
            total_sampled_mass = 0
            multiplier = 1
            a_0 = rndm(a=0.08, b=150, g=-1.35, size=size*multiplier)

            total_sampled_mass += np.sum(a_0)

            return a_0, total_sampled_mass

    # sample secondary mass
    def sample_secondary(self, primary_mass):
        """Sample a secondary mass using draws from a uniform mass ratio distribution motivated by
        `Mazeh et al. (1992) <http://adsabs.harvard.edu/abs/1992ApJ...401..265M>`_
        and `Goldberg & Mazeh (1994) <http://adsabs.harvard.edu/abs/1994ApJ...429..362G>`_

        Parameters
        ----------
        primary_mass : array
            sets the maximum secondary mass (for a maximum mass ratio of 1)

        Returns
        -------
        secondary_mass : array
            sampled secondary masses with array size matching size of
            primary_mass
        """

        a_0 = np.random.uniform(0.001, 1, primary_mass.size)
        secondary_mass = primary_mass*a_0

        return secondary_mass


    def binary_select(self, primary_mass, binfrac_model=0.5):
        """Select which primary masses will have a companion using
        either a binary fraction specified by a float or a
        primary-mass dependent binary fraction following
        `van Haaften et al.(2009) <http://adsabs.harvard.edu/abs/2013A%26A...552A..69V>`_ in appdx

        Parameters
        ----------
        primary_mass : array
            Mass that determines the binary fraction
        binfrac_model : str or float
            vanHaaften - primary mass dependent and ONLY VALID
                         up to 100 Msun
            float - fraction of binaries; 0.5 means 2 in 3 stars are a binary pair while 1 means every star is in a binary pair

        Returns
        -------
        primary_mass[binaryIdx] : array
            primary masses that will have a binary companion
        primary_mass[singleIdx] : array
            primary masses that will be single stars
        binary_fraction[binaryIdx] : array
            system-specific probability of being in a binary
        """

        if type(binfrac_model) == str:
            if binfrac_model == 'vanHaaften':
                binary_fraction = 1/2.0 + 1/4.0 * np.log10(primary_mass)
                binary_choose =  np.random.uniform(0, 1.0, primary_mass.size)

                singleIdx, = np.where(binary_fraction < binary_choose)
                binaryIdx, = np.where(binary_fraction >= binary_choose)
            else:
                raise ValueError('You have supplied a non-supported binary fraction model. Please choose vanHaaften or a float')
        elif type(binfrac_model) == float:
            if (binfrac_model <= 1.0) & (binfrac_model >= 0.0):
                binary_fraction = binfrac_model * np.ones(primary_mass.size)
                binary_choose = np.random.uniform(0, 1.0, primary_mass.size)

                singleIdx, = np.where(binary_choose > binary_fraction)
                binaryIdx, = np.where(binary_choose <= binary_fraction)
            else:
                raise ValueError('You have supplied a fraction outside of 0-1. Please choose a fraction between 0 and 1.')
        else:
            raise ValueError('You have not supplied a model or a fraction. Please choose either vanHaaften or a float')

        return primary_mass[binaryIdx], primary_mass[singleIdx], binary_fraction[binaryIdx]


    def sample_porb(self, mass1, mass2, ecc, size=None):
        """Sample the semi-major axis flat in log space from RROL < 0.5 up
        to 1e5 Rsun according to
        `Abt (1983) <http://adsabs.harvard.edu/abs/1983ARA%26A..21..343A>`_
        and consistent with Dominik+2012,2013
        and then converted to orbital period in days using Kepler III

        Parameters
        ----------
        mass1 : array
            primary masses
        mass2 : array
            secondary masses
        ecc : array
            eccentricities

        Returns
        -------
        porb_sec/sec_in_day : array
            orbital period with array size equalling array size
            of mass1 and mass2
        """
        q = mass2/mass1
        RL_fac = (0.49*q**(2./3.)) / (0.6*q**(2./3.) + np.log(1+q**1./3.))

        q2 = mass1/mass2
        RL_fac2 = (0.49*q2**(2./3.)) / (0.6*q2**(2./3.) + np.log(1+q2**1./3.))
        try:
            ind_lo, = np.where(mass1 < 1.66)
            ind_hi, = np.where(mass1 >= 1.66)

            rad1 = np.zeros(len(mass1))
            rad1[ind_lo] = 1.06*mass1[ind_lo]**0.945
            rad1[ind_hi] = 1.33*mass1[ind_hi]**0.555
        except:
            if mass1 < 1.66:
                rad1 = 1.06*mass1**0.945
            else:
                rad1 = 1.33*mass1**0.555

        try:
            ind_lo, = np.where(mass2 < 1.66)
            ind_hi, = np.where(mass2 >= 1.66)

            rad2 = np.zeros(len(mass2))
            rad2[ind_lo] = 1.06*mass2[ind_lo]**0.945
            rad2[ind_hi] = 1.33*mass2[ind_hi]**0.555
        except:
            if mass2 < 1.66:
                rad2 = 1.06*mass1**0.945
            else:
                rad2 = 1.33*mass1**0.555

        # include the factor for the eccentricity
        RL_max = 2*rad1/RL_fac
        ind_switch, = np.where(RL_max < 2*rad2/RL_fac2)
        if len(ind_switch) >= 1:
            RL_max[ind_switch] = 2*rad2/RL_fac2[ind_switch]
        a_min = RL_max*(1+ecc)
        a_0 = np.random.uniform(np.log(a_min), np.log(1e5), size)

        # convert out of log space
        a_0 = np.exp(a_0)
        # convert to au
        rsun_au = 0.00465047
        a_0 = a_0*rsun_au

        # convert to orbital period in years
        yr_day = 365.24
        porb_yr = ((a_0**3.0)/(mass1+mass2))**0.5
        return porb_yr*yr_day


    def sample_ecc(self, ecc_model='thermal', size=None):
        """Sample the eccentricity according to a user specified model

        Parameters
        ----------
        ecc_model : string
            'thermal' samples from a  thermal eccentricity distribution following
            `Heggie (1975) <http://adsabs.harvard.edu/abs/1975MNRAS.173..729H>`_
            'uniform' samples from a uniform eccentricity distribution
            DEFAULT = 'thermal'

        size : int, optional
            number of eccentricities to sample
            NOTE: this is set in cosmic-pop call as Nstep

        Returns
        -------
        ecc : array
            array of sampled eccentricities with size=size
        """

        if ecc_model=='thermal':
            a_0 = np.random.uniform(0.0, 1.0, size)
            ecc = a_0**0.5
            return ecc

        elif ecc_model=='uniform':
            ecc = np.random.uniform(0.0, 1.0, size)
            return ecc

        else:
            raise Error('You have specified an unsupported model. Please choose from thermal or uniform')

    def sample_SFH(self, SFH_model='const', component_age=10000.0, met=0.02, size=None):
        """Sample an evolution time for each binary based on a user-specified
        star formation history (SFH) and Galactic component age.
        The default is a MW thin disk constant evolution over 10000 Myr

        Parameters
        ----------
        SFH_model : str
            'const' assigns an evolution time assuming a constant star
            formation rate over the age of the MW disk: component_age [Myr]
            'burst' assigns an evolution time assuming a burst of constant
            star formation for 1Gyr starting at component_age [Myr] in the past
            'delta_burst' assignes a t=0 evolution time until component age
            DEFAULT: 'const'
        component_age: float
            age of the Galactic component [Myr]; DEFAULT: 10000.0
        met : float
            metallicity of the population [Z_sun = 0.02]
            Default: 0.02
        size : int, optional
            number of evolution times to sample
            NOTE: this is set in cosmic-pop call as Nstep

        Returns
        -------
        tphys : array
            array of evolution times of size=size
        metallicity : array
            array of metallicities
        """

        if SFH_model=='const':

            tphys = np.random.uniform(0, component_age, size)
            metallicity = np.ones(size)*met
            return tphys, metallicity

        elif SFH_model=='burst':
            tphys = component_age - np.random.uniform(0, 1000, size)
            metallicity = np.ones(size)*met
            return tphys, metallicity

        elif SFH_model=='delta_burst':
            tphys = component_age*np.ones(size)
            metallicity = np.ones(size)*met
            return tphys, metallicity

        else:
            raise Error('You have specified an unsupported model. Please choose from const, burst, or delta_burst')

    def set_kstar(self, mass):
        """Initialize stellar types according to BSE classification
        kstar=1 if M>=0.7 Msun; kstar=0 if M<0.7 Msun

        Parameters
        ----------
        mass : array
            array of masses

        Returns
        -------
        kstar : array
            array of initial stellar types
        """

        kstar = np.zeros(mass.size)
        low_cutoff = 0.7
        lowIdx = np.where(mass < low_cutoff)[0]
        hiIdx = np.where(mass >= low_cutoff)[0]

        kstar[lowIdx] = 0
        kstar[hiIdx] = 1

        return kstar




# -*- coding: utf-8 -*-
# Copyright (C) Katelyn Breivik (2017 - 2019)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

'''`plotting`
'''

import numpy as np
import pandas as pd
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.patches as mpatches
from matplotlib.lines import Line2D
from .utils import a_from_p, p_from_a, calc_Roche_radius
from .evolve import Evolve


__author__ = 'Jeff Andrews <andrews.jeff@gmail.com>'
__all__ = ['evolve_binary', 'plot_k_type', 'plot_radius',
           'plot_mass','plot_Teff','plot_Mdot','plot_P_orb','plot_ecc',
           'plot_HR_diagram','plot_binary_evol','evolve_and_plot']


rsun_in_au = 215.0954
day_in_year = 365.242


# Colors
primary_color = 'C0'
secondary_color = 'C1'


def evolve_binary(initC, t_min=None, t_max=None, BSEDict={}):
    """
    Evolves a single binary with all timesteps written
    to the bcm array for plotting

    Parameters
    ----------
    initC : `pandas.DataFrame`
        initial conditions for binary to evolve

    t_min : `float`
        starting time for plot in Myr

    t_max : `float`
        ending time for plot in Myr

    BSEDict : `Dict
        Dictionary containing all BSE flags needed

    Returns
    -------
    bcm : `pandas.DataFrame`
        binary evolution data form BSE's bcm dict
    """

    # Disable chained warning for now which throws
    # a warning for setting dtp and tphysf
    pd.options.mode.chained_assignment = None

    # Get highest BSE temporal resolution
    initC['dtp'] = 0.01

    # To deal with the limited size of the bcm array, we need to reduce
    # the time resolution if we are evolving a binary for more than 100 Myr
    if not t_max is None:
        if not t_min is None:
            if t_max-t_min > 100: initC['dtp'] = (t_max-t_min) / 10000
        else:
            if t_max > 100: initC['dtp'] = t_max / 10000

    # Set maximum time for evolution
    if not t_max is None: initC['tphysf'] = t_max

    # Call evolution scripts
    bpp, bcm, initC = Evolve.evolve(initialbinarytable=initC, BSEDict=BSEDict)

    # In case minimum and maximum times are not supplied by user
    if t_min is None:
        # BUG HERE - BSE sets first value to weird float
        # HACKED FIX
        t_phys = np.array(bcm.tphys)
        t_phys[0] = 0.0
        bcm.at[0, 'tphys'] = t_phys
        t_min = bcm.tphys.min()
    if t_max is None: t_max = bcm.tphys.max()

    # Select for temporal range
    bcm = bcm.loc[(bcm.tphys<=t_max) & (bcm.tphys>=t_min)]

    return bcm


def plot_k_type(ax_1, ax_2, ax_k_type_list, times, k1_out, k2_out):
    """Plots the stellar types as a function of time

    Parameters
    ----------
    ax_1 : `matplotlib.Axes`
        Axis instance for kstar 1

    ax_2 : `matplotlib.Axes`
        Axis instance for kstar 2

    ax_k_type_list : `list`
        List of ktypes for the legend of the ktype bar plots

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    k1_out : `pandas.Series`
        Series of kstar 1 type at each binary evolution time step in Myr

    k2_out : `pandas.Series`
        Series of kstar 2 type at each binary evolution time step in Myr

    Returns
    -------
    Null
    """

    k_type_colors = ['plum', 'sandybrown', 'lightseagreen', 'moccasin', 'chartreuse',
                     'deepskyblue', 'gold', 'rosybrown', 'm', 'darkgreen', 'grey',
                     'sienna', 'palevioletred', 'navy', 'tan', 'black']
    k_type = ['MS conv', 'MS', 'HG', 'GB', 'CHeB', 'EAGB', 'TPAGB', 'HeMS', 'HeHG',
              'HeGB', 'HeWD', 'COWD', 'ONeWD', 'NS', 'BH', 'no remnant']

    # k-type plots
    for a in [ax_1, ax_2]:

        a.axis('off')

        for j,k in enumerate(np.unique(k1_out)):
            a.fill_between(times[k1_out == k], 0.7, 0.9, color=k_type_colors[k])
        for j,k in enumerate(np.unique(k2_out)):
            a.fill_between(times[k2_out == k], 0.4, 0.6, color=k_type_colors[k])

        a.set_title('Stellar Type')

        a.set_xlim(np.min(times), np.max(times))



    # Add legend
    ax_k_type_list.axis('off')
    k_type_all = np.unique(np.array([k1_out, k2_out]))
    patches = [ mpatches.Patch(color=k_type_colors[k], label=k_type[k]) for k in k_type_all ]
    leg = ax_k_type_list.legend(handles=patches, mode='expand', ncol=len(k_type_all))
    leg.get_frame().set_alpha(0.0)



def plot_radius(ax, times, R1_out, R2_out, M1_out, M2_out, P_orb_out, ecc_out, sys_obs, legend=True):
    """Plots the stellar radii as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    R1_out : `pandas.Series`
        Series of primary radius at each binary evolution time step in Rsun

    R2_out : `pandas.Series`
        Series of secondary radius at each binary evolution time step in Rsun

    M1_out : `pandas.Series`
        Series of primary mass at each binary evolution time step in Msun

    M2_out : `pandas.Series`
        Series of secondary mass at each binary evolution time step in Msun

    P_orb_out : `pandas.Series`
        Series of orbital periods at each binary evolution time step in days

    ecc_out : `pandas.Series`
        Series of eccentricities at each binary evolution time step

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """
    # Radius
    ax.plot(times, R1_out, color=primary_color)
    ax.plot(times, R2_out, color=secondary_color)

    # Plot Roche radii - at periastron
    A_out = a_from_p(M1_out, M2_out, P_orb_out) # in Rsun
    R1_Roche = calc_Roche_radius(M1_out, M2_out, A_out*(1.0-ecc_out)) # in Rsun
    R2_Roche = calc_Roche_radius(M2_out, M1_out, A_out*(1.0-ecc_out)) # in Rsun
    ax.plot(times, R1_Roche, color=primary_color, linestyle='--')
    ax.plot(times, R2_Roche, color=secondary_color, linestyle='--')


    for key, value in sys_obs.items():
        if key == 'R1': ax.axhline(value, color=primary_color, linestyle='dashed')
        if key == 'R2': ax.axhline(value, color=secondary_color, linestyle='dashed')


    ax.set_yscale('log')
    ax.set_ylim(0.5, ax.get_ylim()[1])

    ax.set_ylabel(r'Radius ($R_{\odot}$)')

    ax.set_xlim(np.min(times), np.max(times))

    if legend:
        custom_lines = [Line2D([0], [0], color='k', linestyle='solid'),
                        Line2D([0], [0], color='k', linestyle='dashed')]
        ax.legend(custom_lines, ['Stellar Radius', 'Roche Radius'], frameon=False)


def plot_mass(ax, times, M1_out, M2_out, sys_obs):
    """Plots the stellar mass as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    M1_out : `pandas.Series`
        Series of primary mass at each binary evolution time step in Msun

    M2_out : `pandas.Series`
        Series of secondary mass at each binary evolution time step in Msun

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    # Mass
    ax.plot(times, M1_out, color=primary_color)
    ax.plot(times, M2_out, color=secondary_color)

    for key, value in sys_obs.items():
        if key == 'M1': ax.axhline(value, color=primary_color, linestyle='dashed')
        if key == 'M2': ax.axhline(value, color=secondary_color, linestyle='dashed')

    ax.set_ylabel(r'Mass ($M_{\odot}$)')
    ax.set_xlim(np.min(times), np.max(times))



def plot_Teff(ax, times, Teff1_out, Teff2_out, sys_obs):
    """Plots the stellar effectvie temperature as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    Teff1_out : `pandas.Series`
        Series of primary effective temperature at each binary evolution time step in K

    Teff2_out : `pandas.Series`
        Series of secondary effective temperature at each binary evolution time step in K

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    # Effective temperature
    ax.plot(times, Teff1_out, color=primary_color)
    ax.plot(times, Teff2_out, color=secondary_color)

    for key, value in sys_obs.items():
        if key == 'T1': ax.axhline(value, color=primary_color, linestyle='dashed')
        if key == 'T2': ax.axhline(value, color=secondary_color, linestyle='dashed')

    ax.set_yscale('log')
    ax.set_ylabel(r'T$_{\rm eff}$ (K)')
    ax.set_xlim(np.min(times), np.max(times))
    ax.set_xlabel("Time (Myr)")


def plot_Mdot(ax, times, Mdot1_out, Mdot2_out, legend=True):
    """Plots the stellar effectvie temperature as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    Mdot1_out : `pandas.Series`
        Series of primary mass transfer rate at each binary evolution time step in Msun/yr

    Mdot2_out : `pandas.Series`
        Series of secondary mass transfer rate at each binary evolution time step in Msun/yr

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """


    # Mass accretion rate
    ax.plot(times, np.log10(np.clip(Mdot1_out, 1.0e-16, None)), color=primary_color)
    ax.plot(times, np.log10(np.clip(Mdot2_out, 1.0e-16, None)), color=secondary_color)
    ax.set_ylabel(r'Mass Accretion Rate ($M_{\odot}$ yr$^{-1}$)')

    ax.set_ylim(-14, ax.get_ylim()[1])
    ax.set_xlim(np.min(times), np.max(times))

    if legend:
        custom_lines = [Line2D([0], [0], color='C0', linestyle='solid'),
                        Line2D([0], [0], color='C1', linestyle='solid')]
        ax.legend(custom_lines, ['Primary', 'Secondary'], loc=2, frameon=False)



def plot_P_orb(ax, times, P_orb_out, t_max, sys_obs):
    """Plots the orbital period as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    P_orb_out : `pandas.Series`
        Series of orbital periods at each binary evolution time step in days

    t_max : `float`
        ending time for plot in Myr

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    ax.plot(times, P_orb_out*day_in_year, color='k')

    for key, value in sys_obs.items():
        if key == 'P_orb': ax.axhline(value, color='k', linestyle='dashed')

    ax.set_xlim(np.min(times), t_max)

    ax.set_ylabel(r'P$_{\rm orb}$ (days)')
    ax.set_yscale('log')


def plot_ecc(ax, times, ecc_out, t_max, sys_obs):
    """Plots the eccentricity as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    ecc_out : `pandas.Series`
        Series of eccentricities at each binary evolution time step

    t_max : `float`
        ending time for plot in Myr

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    ax.plot(times, ecc_out, color='k')

    for key, value in sys_obs.items():
        if key == 'ecc': ax.axhline(value, color='k', linestyle='dashed')


    ax.set_xlim(np.min(times), t_max)

    ax.set_ylabel('Eccentricity')
    ax.set_xlabel("Time (Myr)")


def plot_HR_diagram(ax, L1_out, L2_out, Teff1_out, Teff2_out):
    """Plots the eccentricity as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    L1_out : `pandas.Series`
        Series of primary luminosities at each binary evolution time step in Lsun

    L2_out : `pandas.Series`
        Series of secondary luminosities at each binary evolution time step in Lsun

    Teff1_out : `pandas.Series`
        Series of primary effective temperature at each binary evolution time step in K

    Teff2_out : `pandas.Series`
        Series of secondary effective temperature at each binary evolution time step in K

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    ax.plot(Teff1_out, L1_out, color=primary_color)
    ax.plot(Teff2_out, L2_out, color=secondary_color)

    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.set_xlim(40000, 2000)

    ax.set_xlabel(r"log T$_{\rm eff}$ (K)")
    ax.set_ylabel("log L (erg/s)")

    ax.set_xticks([2000, 5000, 10000, 20000, 40000])



def plot_binary_evol(bcm, sys_obs={}):
    """Plots the full set of plots and kstar bar plots

    Parameters
    ----------
    bcm : `pandas.DataFrame`
        binary evolution data form BSE's bcm dict

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    fig : `matplotlib.figure`
        Figure containing all the plots!
    """

    fig, ax = plt.subplots(5, 1, figsize=(10,10))

    gs = gridspec.GridSpec(4, 2,
                           width_ratios=[2,2],
                           height_ratios=[1,2,2,2]
                           )

    ax = np.array([plt.subplot(x) for x in gs])


    # k-type panels
    ax_k_type_list = fig.add_axes([0.08, 0.75, 0.9, 0.1])
    plot_k_type(ax[0], ax[1], ax_k_type_list, bcm.tphys, bcm.kstar_1.astype(int), bcm.kstar_2.astype(int))

    # Radius panel
    plot_radius(ax[2], bcm.tphys, bcm.rad_1, bcm.rad_2, bcm.mass_1, bcm.mass_2, bcm.porb, bcm.ecc, sys_obs)

    # Mass panel
    plot_mass(ax[4], bcm.tphys, bcm.mass_1, bcm.mass_2, sys_obs)

    # Teff panel
    plot_Teff(ax[6], bcm.tphys, bcm.teff_1, bcm.teff_2, sys_obs)

    # Mass accretion rate panel
    plot_Mdot(ax[3], bcm.tphys, bcm.deltam_1, bcm.deltam_2)

    # Orbital period panel
    plot_P_orb(ax[5], bcm.loc[bcm.kstar_2<15].tphys, bcm.loc[bcm.kstar_2<15].porb, np.max(bcm.tphys), sys_obs)

    # Plot eccentricity panel
    plot_ecc(ax[7], bcm.loc[bcm.kstar_2<15].tphys, bcm.loc[bcm.kstar_2<15].ecc, np.max(bcm.tphys), sys_obs)


    # Plot HR diagram
    # idx_1 = np.where(k1_out < 10)[0]
    # idx_2 = np.where(k2_out < 10)[0]
    # plot_HR_diagram(ax[7], L1_out[k1_out<10], L2_out[k2_out<10], Teff1_out[k1_out<10], Teff2_out[k2_out<10])

    # make the labels look nice
    gs.tight_layout(fig)

    return fig


def evolve_and_plot(initC, t_min=None, t_max=None, BSEDict=None, sys_obs={}):
    """
    Evolve and plot binaries as a function of time

    Parameters
    ----------
    initC : `pandas.DataFrame`
        initial conditions for binary to evolve

    t_min : `float or list`
        starting time for plot in Myr

    t_max : `float or list`
        ending time for plot in Myr

    BSEDict : `Dict
        Dictionary containing all BSE flags needed

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    all_figs : `list`
        list of all figures created
    """

    # Throw an error if user is plotting more than 10 binaries
    if len(initC) > 10:
        raise ValueError("You have asked to plot more than 10 separate binaries. This could cause problems...")

    # Iterate over all binaries in initC
    all_figs = []
    for i in range(len(initC)):

        # Check if t_min and t_max are lists
        if isinstance(t_min, list):
            t_min_tmp = t_min[i]
        else:
            t_min_tmp = t_min
        if isinstance(t_max, list):
            t_max_tmp = t_max[i]
        else:
            t_max_tmp = t_max

        # Evolve binary
        bcm = evolve_binary(initC.iloc[i:i+1], t_min=t_min_tmp, t_max=t_max_tmp, BSEDict=BSEDict)

        # Plot binary
        fig = plot_binary_evol(bcm, sys_obs=sys_obs)

        # Add to list of figs
        all_figs.append(fig)

    # Return list of figs
    return all_figs



# -*- coding: utf-8 -*-
# Copyright (C) Katelyn Breivik (2017 - 2019)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

'''`MC_samp`
'''

import numpy as np
import math
import random
import scipy.special as ss
import scipy.stats as stats

__author__ = 'Katelyn Breivik <katie.breivik@gmail.com>'
__credits__ = 'Scott Coughlin <scott.coughlin@ligo.org>'
__all__ = ['mass_weighted_number', 'select_component_mass',
           'sample_sech_squared', 'sample_exponential_radial',
           'sample_exponential_vertical', 'sample_exponential_square_radial',
           'galactic_positions']


G = 6.67384*math.pow(10, -11.0)
c = 2.99792458*math.pow(10, 8.0)
parsec = 3.08567758*math.pow(10, 16)
Rsun = 6.955*math.pow(10, 8)
Msun = 1.9891*math.pow(10,30)
day = 86400.0
rsun_in_au = 215.0954
day_in_year = 365.242
sec_in_day = 86400.0
sec_in_hour = 3600.0
hrs_in_day = 24.0
sec_in_year = 3.15569*10**7.0
Tobs = 3.15569*10**7.0
geo_mass = G/c**2

# solar coordinates in the galaxy: in parsecs from
# (Chaper 8 of Galactic Structure and stellar Pops book) Yoshii (2013)
############################################################################
x_sun = 8.0
y_sun = 0.0
z_sun = 25.0/1000.0

def mass_weighted_number(dat, total_sampled_mass, component_mass):
    """Compute the total number of systems in the synthetic catalog
    based on the total sampled mass of the simulated system and
    the total mass of a given galactic component

    Parameters
    ----------
    dat : DataFrame
        DataFrame containing the fixed population created from cosmic-pop
    total_sampled_mass : float
        total amount of mass sampled to generate the fixed population
        including single stars
    component_mass : float
        mass of the Galactic component we are simulating

    Returns
    -------
    nSystems : int
        number of systems in a Milky Way population for the selected
        Galactic population and fixed population
    """

    nSystems = int(len(dat.mass_1)*component_mass/total_sampled_mass)
    return nSystems

def select_component_mass(gx_component):
    """Select the Galactic component mass according to
    McMillan (2011)

    Parameters
    ----------
    gx_component : str
        Choose from: 'ThinDisk', 'Bulge', 'ThickDisk'

    Returns
    -------
    gx_component_mass : float
        Galactic component mass [Msun]
    """

    if gx_component == 'ThinDisk':
        gx_component_mass = 4.32e10
    elif gx_component == 'Bulge':
        gx_component_mass = 8.9e9
    elif gx_component == 'ThickDisk':
        gx_component_mass = 1.44e10

    return gx_component_mass

def sample_sech_squared(size, scale_height=0.3):
    """Sample a collection of numbers of size=size distributed according
    to a sech sqaured function with a user specific scale height

    Parameters
    ----------
    size : int
        Size of the sample
    scale_height : float
        Scale height of the distribution; Default=0.3

    Returns
    -------
    distributed_nums : array
        Array of sampled values
    """

    rand_nums = np.random.uniform(0, 1, size)
    distributed_nums = np.arctanh( (2*rand_nums-1) ) * scale_height

    return distributed_nums

def sample_exponential_radial(size, scale_height):
    """Sample a collection of numbers of size=size distributed according
    to a radial exponential function with a user specific scale height

    Parameters
    ----------
    size : int
        Size of the sample
    scale_height : float
        Scale height of the distribution

    Returns
    -------
    distributed_nums : array
        Array of sampled values
    """

    rand_nums = np.random.uniform(0, 1, size)
    distributed_nums = -scale_height * np.log(1.0 - rand_nums)

    return distributed_nums

def sample_exponential_vertical(size, scale_height):
    """Sample a collection of numbers of size=size distributed according
    to a vertical exponential function with a user specific scale height

    Parameters
    ----------
    size : int
        Size of the sample
    scale_height : float
        Scale height of the distribution

    Returns
    -------
    distributed_nums : array
        Array of sampled values
    """

    rand_nums = np.random.uniform(0, 1, size)
    distributed_nums = -scale_height * np.log(1.0 - rand_nums)

    # CHOOSE A POSITION ABOVE AND BELOW THE DISK
    ###########################################################################
    pos_neg_choose = np.random.uniform(0, 1, size)
    negInd, = np.where(pos_neg_choose < 0.5)
    distributed_nums[negInd] = distributed_nums[negInd] * (-1.0)

    return distributed_nums

def sample_exponential_square_radial(size, scale_height):
    """Sample a collection of numbers of size=size distributed according
    to an exponential squared function with a user specific scale height

    Parameters
    ----------
    size : int
        Size of the sample
    scale_height : float
        Scale height of the distribution

    Returns
    -------
    distributed_nums : array
        Array of sampled values
    """

    rand_nums = np.random.uniform(0, 1, size)
    distributed_nums = scale_height * ss.erfinv(rand_nums)

    return distributed_nums

def galactic_positions(gx_component, size, model='McMillan'):
    """Sample a set of Galactic positions of size=size distributed
    according to the user specified model. X,Y,Z positions in [pc];
    Galactocentric distance in [kpc]

    Parameters
    ----------
    gx_component : str
        Choose from : 'ThinDisk', 'Bulge', 'ThickDisk'
    size : int
        Size of the sample
    model : str
        Current default model is 'McMillan'

    Returns
    -------
    xGx, yGx, zGx, inc, OMEGA, omega : array
        Array of sampled positions in Galactic cartesian coordinates
        centered on the Galactic center and orientations in radians
    """
    # SAMPLE FROM VERY SIMPLE DISTRIBUTION FUNCTIONS
    ###########################################################################
    if gx_component == 'ThinDisk':
        # COMPUTE THE POSITION AND ORIENTATION OF EACH BINARY
        #######################################################################
        # Assign the radial and vertical positions:
        if model == 'double_exp':
            r = sample_exponential_radial(size, 2.5)
            z = sample_exponential_vertical(size, 0.3)

        if model == 'sech_squared':
            r = sample_exponential_radial(size, 2.5)
            z = sample_sech_squared(size, 0.3)

        if model == 'McMillan':
            r = sample_exponential_radial(size, 2.9)
            z = sample_exponential_vertical(size, 0.3)

        # Assign the azimuthal positions:
        phi = np.random.uniform(0, 2*np.pi, size)

        # convert to cartesian:
        xGX = r * np.cos(phi)
        yGX = r * np.sin(phi)
        zGX = z

    elif gx_component == 'Bulge':

        # COMPUTE THE POSITION AND ORIENTATION OF EACH BINARY
        #######################################################################
        # Assign the radial positions:
        if model == 'exp_squared':
            r = sample_exponential_square_radial(size, 0.5)
            # Assign the polar positions (uniform in cos(theta):
            theta = np.pi - np.arccos(np.random.uniform(-1, 1, size))

            # Assign the azimuthal positions:
            phi = np.random.uniform(0, 2*np.pi , size)

            # convert to cartesian:
            xGX = r * np.cos(phi) * np.sin(theta)
            yGX = r * np.sin(phi) * np.sin(theta)
            zGX = r * np.cos(theta)

        elif model == 'McMillan':
            r_save = []
            z_save = []
            # sample double exp func and then rejection sample
            while len(r_save) < size:
                rcut = 2.1
                q = 0.5
                r0 = 0.075
                alpha = -1.8
                r = np.random.uniform(0,5,size*10)
                z = np.random.uniform(0,3,size*10)
                prob = np.random.uniform(0,1,size*10)
                sample_func = np.exp(-(r**2 + (z/q)**2)/rcut**2)
                actual_func = (1+np.sqrt((r**2 + (z/q)**2))/r0)**(alpha)*sample_func
                indSave, = np.where(prob < actual_func)
                for ii in indSave:
                    r_save.append(r[ii])
                    z_save.append(z[ii])
            r = np.array(r_save[:size])
            z = np.array(z_save[:size])
            ind_pos_neg = np.random.uniform(0,1,len(z))
            ind_negative, = np.where(ind_pos_neg > 0.5)
            z[ind_negative] = -z[ind_negative]

            # Assign the azimuthal positions:
            phi = np.random.uniform(0, 2*np.pi , size)

            # convert to cartesian:
            xGX = r * np.cos(phi)
            yGX = r * np.sin(phi)
            zGX = z

    elif gx_component == 'ThickDisk':
        # COMPUTE THE POSITION AND ORIENTATION OF EACH BINARY
        ###############################################################################
        # Assign the radial and vertical positions:
        if model == 'double_exp':
            r = sample_exponential_radial(size, 2.5)
            z = sample_exponential_vertical(size, 1.0)

        if model == 'McMillan':
            r = sample_exponential_radial(size, 3.31)
            z = sample_exponential_vertical(size, 0.9)

        # Assign the azimuthal position of the star
        phi = np.random.uniform(0, 2*np.pi, size)

        # convert to cartesian:
        xGX = r*np.cos(phi)
        yGX = r*np.sin(phi)
        zGX = z

    # assign an inclination, argument of periapsis, and longitude of ascending node
    inc = np.pi - np.arccos(np.random.uniform(-1, 0, size))
    OMEGA = np.random.uniform(0, 2*np.pi, size)
    omega = np.random.uniform(0, 2*np.pi, size)

    return xGX, yGX, zGX, inc, OMEGA, omega



# -*- coding: utf-8 -*-
# Copyright (C) Scott Coughlin (2017 - 2019)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from . import zdata

__author__ = 'Scott Coughlin <scott.coughlin@ligo.org>'
__all__ = ['zcnsts']

c = numpy.array([3.040581e-01, 8.049509e-02, 8.967485e-02, 8.780198e-02, 2.219170e-02])

def zcnsts(z):
    """
    *
    *       ------------------------------------------------------------
    *
    *      zpars:  1; M below which hook doesn't appear on MS, Mhook.
    *              2; M above which He ignition occurs non-degenerately, Mhef.
    *              3; M above which He ignition occurs on the HG, Mfgb.
    *              4; M below which C/O ignition doesn't occur, Mup.
    *              5; M above which C ignites in the centre, Mec.
    *              6; value of log D for M<= zpars[2]
    *              7; value of x for Rgb propto M^(-x)
    *              8; value of x for tMS = numpy.maximum(tHOOK,x*tBGB)
    *              9; constant for McHeIf when computing Mc,BGB, mchefl.
    *             10; constant for McHeIf when computing Mc,HeI, mchefl.
    *             11; hydrogen abundance.
    *             12; helium abundance.
    *             13; constant x in rnumpy.minimum = rgb*x**y used by LM CHeB.
    *             14; z**0.4 to be used for WD L formula.
    *
    *       ------------------------------------------------------------
    *
    """

    # initialize arrays
    msp = numpy.zeros(shape=(200, len(z)))
    gbp = numpy.zeros(shape=(200, len(z)))
    zpars = numpy.zeros(shape=(20, len(z)))
    tscls = numpy.zeros(shape=(20, len(z)))
    lums = numpy.zeros(shape=(10, len(z)))
    GB = numpy.zeros(shape=(10, len(z)))

    lzs = numpy.log10(z/0.020)
    dlzs = 1.0/(z*numpy.log(10.0))
    lz = numpy.log10(z)
    lzd = lzs + 1.0

    zpars[0] = 1.01850 + lzs*(0.16015 + lzs*0.0892)
    zpars[1] = 1.9950 + lzs*(0.25 + lzs*0.087)
    zpars[2] = 16.50*z**0.06/(1.0 + (1.0e-04/z)**1.27)
    zpars[3] = numpy.maximum(6.110440 + 1.02167*lzs, 5.0)
    zpars[4] = zpars[3] + 1.80
    zpars[5] = 5.370 + lzs*0.135
    zpars[6] = c[0] + lzs*(c[1] + lzs*(c[2] + lzs*(c[3] + lzs*c[4])))
    zpars[7] = numpy.maximum(0.950,numpy.maximum(0.95-(10.0/3.0)*(z-0.01),
               numpy.minimum(0.990,0.98-(100.0/7.0)*(z-0.001))))

    # Lzams

    msp[0] = zdata.xz[0]+lzs*(zdata.xz[1]+lzs*(zdata.xz[2]+lzs*(zdata.xz[3]+lzs*zdata.xz[4])))
    msp[1] = zdata.xz[5]+lzs*(zdata.xz[6]+lzs*(zdata.xz[7]+lzs*(zdata.xz[8]+lzs*zdata.xz[9])))
    msp[2] = zdata.xz[10]+lzs*(zdata.xz[11]+lzs*(zdata.xz[12]+lzs*(zdata.xz[13]+lzs*zdata.xz[14])))
    msp[3] = zdata.xz[15]+lzs*(zdata.xz[16]+lzs*(zdata.xz[17]+lzs*(zdata.xz[18]+lzs*zdata.xz[19])))
    msp[4] = zdata.xz[20]+lzs*(zdata.xz[21]+lzs*(zdata.xz[22]+lzs*(zdata.xz[23]+lzs*zdata.xz[24])))
    msp[5] = zdata.xz[25]+lzs*(zdata.xz[26]+lzs*(zdata.xz[27]+lzs*(zdata.xz[28]+lzs*zdata.xz[29])))
    msp[6] = zdata.xz[30]+lzs*(zdata.xz[31]+lzs*(zdata.xz[32]+lzs*(zdata.xz[33]+lzs*zdata.xz[34])))

    # Rzams

    msp[7] = zdata.xz[35]+lzs*(zdata.xz[36]+lzs*(zdata.xz[37]+lzs*(zdata.xz[38]+lzs*zdata.xz[39])))
    msp[8] = zdata.xz[40]+lzs*(zdata.xz[41]+lzs*(zdata.xz[42]+lzs*(zdata.xz[43]+lzs*zdata.xz[44])))
    msp[9] = zdata.xz[45]+lzs*(zdata.xz[46]+lzs*(zdata.xz[47]+lzs*(zdata.xz[48]+lzs*zdata.xz[49])))
    msp[10] = zdata.xz[50]+lzs*(zdata.xz[51]+lzs*(zdata.xz[52]+lzs*(zdata.xz[53]+lzs*zdata.xz[54])))
    msp[11] = zdata.xz[55]+lzs*(zdata.xz[56]+lzs*(zdata.xz[57]+lzs*(zdata.xz[58]+lzs*zdata.xz[59])))
    msp[12] = zdata.xz[60]
    msp[13] = zdata.xz[61]+lzs*(zdata.xz[62]+lzs*(zdata.xz[63]+lzs*(zdata.xz[64]+lzs*zdata.xz[65])))
    msp[14] = zdata.xz[66]+lzs*(zdata.xz[67]+lzs*(zdata.xz[68]+lzs*(zdata.xz[69]+lzs*zdata.xz[70])))
    msp[15] = zdata.xz[71]+lzs*(zdata.xz[72]+lzs*(zdata.xz[73]+lzs*(zdata.xz[74]+lzs*zdata.xz[75])))

    # Tbgb

    msp[16] = zdata.xt[0]+lzs*(zdata.xt[1]+lzs*(zdata.xt[2]+lzs*zdata.xt[3]))
    msp[17] = zdata.xt[4]+lzs*(zdata.xt[5]+lzs*(zdata.xt[6]+lzs*zdata.xt[7]))
    msp[18] = zdata.xt[8]+lzs*(zdata.xt[9]+lzs*(zdata.xt[10]+lzs*zdata.xt[11]))
    msp[19] = zdata.xt[12]+lzs*(zdata.xt[13]+lzs*(zdata.xt[14]+lzs*zdata.xt[15]))
    msp[20] = zdata.xt[16]

    # dTbgb/dz
    msp[116] = dlzs*(zdata.xt[1]+lzs*(2.0*zdata.xt[2]+3.0*lzs*zdata.xt[3]))
    msp[117] = dlzs*(zdata.xt[5]+lzs*(2.0*zdata.xt[6]+3.0*lzs*zdata.xt[7]))
    msp[118] = dlzs*(zdata.xt[9]+lzs*(2.0*zdata.xt[10]+3.0*lzs*zdata.xt[11]))
    msp[119] = dlzs*(zdata.xt[13]+lzs*(2.0*zdata.xt[14]+3.0*lzs*zdata.xt[15]))

    # Thook
    msp[21] = zdata.xt[17]+lzs*(zdata.xt[18]+lzs*(zdata.xt[19]+lzs*zdata.xt[20]))
    msp[22] = zdata.xt[21]
    msp[23] = zdata.xt[22]+lzs*(zdata.xt[23]+lzs*(zdata.xt[24]+lzs*zdata.xt[25]))
    msp[24] = zdata.xt[26]+lzs*(zdata.xt[27]+lzs*(zdata.xt[28]+lzs*zdata.xt[29]))
    msp[25] = zdata.xt[30]

    # Ltms
    msp[26] = zdata.xl[0]+lzs*(zdata.xl[1]+lzs*(zdata.xl[2]+lzs*(zdata.xl[3]+lzs*zdata.xl[4])))
    msp[27] = zdata.xl[5]+lzs*(zdata.xl[6]+lzs*(zdata.xl[7]+lzs*(zdata.xl[8]+lzs*zdata.xl[9])))
    msp[28] = zdata.xl[10]+lzs*(zdata.xl[11]+lzs*(zdata.xl[12]+lzs*zdata.xl[13]))
    msp[29] = zdata.xl[14]+lzs*(zdata.xl[15]+lzs*(zdata.xl[16]+lzs*(zdata.xl[17]+lzs*zdata.xl[18])))
    msp[26] = msp[26]*msp[29]
    msp[27] = msp[27]*msp[29]
    msp[30] = zdata.xl[19]+lzs*(zdata.xl[20]+lzs*(zdata.xl[21]+lzs*zdata.xl[22]))
    msp[31] = zdata.xl[23]+lzs*(zdata.xl[24]+lzs*(zdata.xl[25]+lzs*zdata.xl[26]))

    # Lalpha
    m2 = 2.0
    msp[32] = zdata.xl[27]+lzs*(zdata.xl[28]+lzs*(zdata.xl[29]+lzs*zdata.xl[30]))
    msp[33] = zdata.xl[31]+lzs*(zdata.xl[32]+lzs*(zdata.xl[33]+lzs*zdata.xl[34]))
    msp[34] = zdata.xl[35]+lzs*(zdata.xl[36]+lzs*(zdata.xl[37]+lzs*zdata.xl[38]))
    msp[35] = zdata.xl[39]+lzs*(zdata.xl[40]+lzs*(zdata.xl[41]+lzs*zdata.xl[42]))
    msp[36] = numpy.maximum(0.90,1.1064+lzs*(0.415+0.18*lzs))
    msp[37] = numpy.maximum(1.0,1.19+lzs*(0.377+0.176*lzs))

    if (z > 0.010).any():
        msp[36, z > 0.010] = numpy.minimum(msp[36, z > 0.010],1.0)
        msp[37, z > 0.010] = numpy.minimum(msp[37, z > 0.010],1.10)

    msp[38] = numpy.maximum(0.1450,0.0977-lzs*(0.231+0.0753*lzs))
    msp[39] = numpy.minimum(0.240+lzs*(0.18+0.595*lzs),0.306+0.053*lzs)
    msp[40] = numpy.minimum(0.330+lzs*(0.132+0.218*lzs),
                  0.36250+0.062*lzs)
    msp[41] = (msp[32]+msp[33]*m2**msp[35])/(m2**0.40+msp[34]*m2**1.9)

    # Lbeta
    msp[42] = zdata.xl[43]+lzs*(zdata.xl[44]+lzs*(zdata.xl[45]+lzs*(zdata.xl[46]+lzs*zdata.xl[47])))
    msp[43] = zdata.xl[48]+lzs*(zdata.xl[49]+lzs*(zdata.xl[50]+lzs*(zdata.xl[51]+lzs*zdata.xl[52])))
    msp[44] = zdata.xl[53]+lzs*(zdata.xl[54]+lzs*zdata.xl[55])
    msp[45] = numpy.minimum(1.40,1.5135+0.3769*lzs)
    msp[45] = numpy.maximum(0.63550-0.4192*lzs,numpy.maximum(1.25,msp[45]))

    # Lhook
    msp[46] = zdata.xl[56]+lzs*(zdata.xl[57]+lzs*(zdata.xl[58]+lzs*zdata.xl[59]))
    msp[47] = zdata.xl[60]+lzs*(zdata.xl[61]+lzs*(zdata.xl[62]+lzs*zdata.xl[63]))
    msp[48] = zdata.xl[64]+lzs*(zdata.xl[65]+lzs*(zdata.xl[66]+lzs*zdata.xl[67]))
    msp[49] = zdata.xl[68]+lzs*(zdata.xl[69]+lzs*(zdata.xl[70]+lzs*zdata.xl[71]))
    msp[50] = numpy.minimum(1.40,1.5135+0.3769*lzs)
    msp[50] = numpy.maximum(0.63550-0.4192*lzs,numpy.maximum(1.25,msp[50]))

    # Rtms
    msp[51] = zdata.xr[0]+lzs*(zdata.xr[1]+lzs*(zdata.xr[2]+lzs*(zdata.xr[3]+lzs*zdata.xr[4])))
    msp[52] = zdata.xr[5]+lzs*(zdata.xr[6]+lzs*(zdata.xr[7]+lzs*(zdata.xr[8]+lzs*zdata.xr[9])))
    msp[53] = zdata.xr[10]+lzs*(zdata.xr[11]+lzs*(zdata.xr[12]+lzs*(zdata.xr[13]+lzs*zdata.xr[14])))
    msp[54] = zdata.xr[15]+lzs*(zdata.xr[16]+lzs*(zdata.xr[17]+lzs*zdata.xr[18]))
    msp[55] = zdata.xr[19]+lzs*(zdata.xr[20]+lzs*(zdata.xr[21]+lzs*zdata.xr[22]))
    msp[51] = msp[51]*msp[53]
    msp[52] = msp[52]*msp[53]
    msp[56] = zdata.xr[23]
    msp[57] = zdata.xr[24]+lzs*(zdata.xr[25]+lzs*(zdata.xr[26]+lzs*zdata.xr[27]))
    msp[58] = zdata.xr[28]+lzs*(zdata.xr[29]+lzs*(zdata.xr[30]+lzs*zdata.xr[31]))
    msp[59] = zdata.xr[32]+lzs*(zdata.xr[33]+lzs*(zdata.xr[34]+lzs*zdata.xr[35]))
    msp[60] = zdata.xr[36]+lzs*(zdata.xr[37]+lzs*(zdata.xr[38]+lzs*zdata.xr[39]))
    #
    msp[61] = numpy.maximum(0.0970-0.1072*(lz+3.0),numpy.maximum(0.097,numpy.minimum(0.1461,
                  0.14610+0.1237*(lz+2.0))))
    msp[61] = 10.0**msp[61]
    m2 = msp[61] + 0.10
    msp[62] = (msp[51]+msp[52]*msp[61]**msp[54])/(msp[53]+msp[61]**msp[55])
    msp[63] = (msp[56]*m2**3+msp[57]*m2**msp[60] + msp[58]*m2**(msp[60]+1.50))/(msp[59]+m2**5)

    # Ralpha
    msp[64] = zdata.xr[40]+lzs*(zdata.xr[41]+lzs*(zdata.xr[42]+lzs*zdata.xr[43]))
    msp[65] = zdata.xr[44]+lzs*(zdata.xr[45]+lzs*(zdata.xr[46]+lzs*zdata.xr[47]))
    msp[66] = zdata.xr[48]+lzs*(zdata.xr[49]+lzs*(zdata.xr[50]+lzs*zdata.xr[51]))
    msp[67] = zdata.xr[52]+lzs*(zdata.xr[53]+lzs*(zdata.xr[54]+lzs*zdata.xr[55]))
    msp[68] = zdata.xr[56]+lzs*(zdata.xr[57]+lzs*(zdata.xr[58]+lzs*(zdata.xr[59]+lzs*zdata.xr[60])))
    msp[69] = numpy.maximum(0.90,numpy.minimum(1.0,1.116+0.166*lzs))
    msp[70] = numpy.maximum(1.4770+0.296*lzs,numpy.minimum(1.6,-0.308-1.046*lzs))
    msp[70] = numpy.maximum(0.80,numpy.minimum(0.8-2.0*lzs,msp[70]))
    msp[71] = zdata.xr[61]+lzs*(zdata.xr[62]+lzs*zdata.xr[63])
    msp[72] = numpy.maximum(0.0650,0.0843-lzs*(0.0475+0.0352*lzs))
    msp[73] = 0.07360+lzs*(0.0749+0.04426*lzs)
    if (z < 0.0040).any():
        msp[73, z < 0.0040] = numpy.minimum(0.055, msp[73, z < 0.0040])

    msp[74] = numpy.maximum(0.0910, numpy.minimum(0.121, 0.136+0.0352*lzs))
    msp[75] = (msp[64]*msp[70]**msp[66])/(msp[65] + msp[70]**msp[67])

    if (msp[69] > msp[70]).any():
        msp[69, msp[69] > msp[70]] = msp[70, msp[69] > msp[70]]
        msp[74, msp[69] > msp[70]] = msp[75, msp[69] > msp[70]]

    # Rbeta
    msp[76] = zdata.xr[64]+lzs*(zdata.xr[65]+lzs*(zdata.xr[66]+lzs*zdata.xr[67]))
    msp[77] = zdata.xr[68]+lzs*(zdata.xr[69]+lzs*(zdata.xr[70]+lzs*zdata.xr[71]))
    msp[78] = zdata.xr[72]+lzs*(zdata.xr[73]+lzs*(zdata.xr[74]+lzs*zdata.xr[75]))
    msp[79] = zdata.xr[76]+lzs*(zdata.xr[77]+lzs*(zdata.xr[78]+lzs*zdata.xr[79]))
    msp[80] = zdata.xr[80]+lzs*(zdata.xr[81]+lzs*lzs*zdata.xr[82])
    if (z > 0.010).any():
        msp[80, z > 0.010] = numpy.maximum(msp[80, z > 0.010], 0.95)

    msp[81] = numpy.maximum(1.40, numpy.minimum(1.6, 1.6+lzs*(0.764+0.3322*lzs)))

    # Rgamma
    msp[82] = numpy.maximum(zdata.xr[83]+lzs*(zdata.xr[84]+lzs*(zdata.xr[85]+lzs*zdata.xr[86])),
                  zdata.xr[95]+lzs*(zdata.xr[96]+lzs*zdata.xr[97]))
    msp[83] = numpy.minimum(0.0, zdata.xr[87]+lzs*(zdata.xr[88]+lzs*(zdata.xr[89]+lzs*zdata.xr[90])))
    msp[83] = numpy.maximum(msp[83], zdata.xr[98]+lzs*(zdata.xr[99]+lzs*zdata.xr[100]))
    msp[84] = zdata.xr[91]+lzs*(zdata.xr[92]+lzs*(zdata.xr[93]+lzs*zdata.xr[94]))
    msp[84] = numpy.maximum(0.0, numpy.minimum(msp[84], 7.454+9.046*lzs))
    msp[85] = numpy.minimum(zdata.xr[101]+lzs*zdata.xr[102], numpy.maximum(2.0, -13.3-18.6*lzs))
    msp[86] = numpy.minimum(1.50,numpy.maximum(0.4,2.493+1.1475*lzs))
    msp[87] = numpy.maximum(1.0,numpy.minimum(1.27,0.8109-0.6282*lzs))
    msp[87] = numpy.maximum(msp[87],0.63550-0.4192*lzs)
    msp[88] = numpy.maximum(5.855420e-02,-0.27110-lzs*(0.5756+0.0838*lzs))

    # Rhook
    msp[89] = zdata.xr[103]+lzs*(zdata.xr[104]+lzs*(zdata.xr[105]+lzs*zdata.xr[106]))
    msp[90] = zdata.xr[107]+lzs*(zdata.xr[108]+lzs*(zdata.xr[109]+lzs*zdata.xr[110]))
    msp[91] = zdata.xr[111]+lzs*(zdata.xr[112]+lzs*(zdata.xr[113]+lzs*zdata.xr[114]))
    msp[92] = zdata.xr[115]+lzs*(zdata.xr[116]+lzs*(zdata.xr[117]+lzs*zdata.xr[118]))
    msp[93] = numpy.minimum(1.250,
              numpy.maximum(1.10,1.9848+lzs*(1.1386+0.3564*lzs)))
    msp[94] = 0.0630 + lzs*(0.0481 + 0.00984*lzs)
    msp[95] = numpy.minimum(1.30,numpy.maximum(0.45,1.2+2.45*lzs))

    # Lneta
    msp[96, z > 0.00090] = 10.0
    msp[96, z < 0.00090] = 20.0

    # converting the below from fortran to python still needs work
    """
    # Lbgb
    gbp[0] = zdata.xg[0]+lzs*(zdata.xg[1]+lzs*(zdata.xg[2]+lzs*zdata.xg[3]))
    gbp[1] = zdata.xg[4]+lzs*(zdata.xg[5]+lzs*(zdata.xg[6]+lzs*zdata.xg[7]))
    gbp[2] = zdata.xg[8]+lzs*(zdata.xg[9]+lzs*(zdata.xg[10]+lzs*zdata.xg[11]))
    gbp[3] = zdata.xg[12]+lzs*(zdata.xg[13]+lzs*(zdata.xg[14]+lzs*zdata.xg[15]))
    gbp[4] = zdata.xg[16]+lzs*(zdata.xg[17]+lzs*zdata.xg[18])
    gbp[5] = zdata.xg[19]+lzs*(zdata.xg[20]+lzs*zdata.xg[21])
    gbp[2] = gbp[2]**gbp[5]
    gbp[6] = zdata.xg[22]
    gbp[7] = zdata.xg[23]

    # Lbagb
    # set gbp[15] = 1.0 until it is reset later with an initial
    # call to Lbagbf using mass = zpars[1] and mhefl = 0.0
    gbp[8] = zdata.xg[24] + lzs*(zdata.xg[25] + lzs*zdata.xg[26])
    gbp[9] = zdata.xg[27] + lzs*(zdata.xg[28] + lzs*zdata.xg[29])
    gbp[10] = 15.0
    gbp[11] = zdata.xg[30]+lzs*(zdata.xg[31]+lzs*(zdata.xg[32]+lzs*zdata.xg[33]))
    gbp[12] = zdata.xg[34]+lzs*(zdata.xg[35]+lzs*(zdata.xg[36]+lzs*zdata.xg[37]))
    gbp[13] = zdata.xg[38]+lzs*(zdata.xg[39]+lzs*(zdata.xg[40]+lzs*zdata.xg[41]))
    gbp[14] = zdata.xg[42]+lzs*zdata.xg[43]
    gbp[11] = gbp[11]**gbp[14]
    gbp[13] = gbp[13]**gbp[14]
    gbp[15] = 1.0

    # Rgb
    gbp[16] = -4.67390-0.9394*lz
    gbp[16] = 10.0**gbp[16]
    gbp[16] = numpy.maximum(gbp[16],-0.041670+55.67*z)
    gbp[16] = numpy.minimum(gbp[16],0.47710-9329.21*z**2.94)
    gbp[17] = numpy.minimum(0.540,0.397+lzs*(0.28826+0.5293*lzs))
    gbp[18] = numpy.maximum(-0.14510,-2.2794-lz*(1.5175+0.254*lz))
    gbp[18] = 10.0**gbp[18]
    if (z > 0.0040):
        gbp[18] = numpy.maximum(gbp[18],0.73070+14265.1*z**3.395)

    gbp[19] = zdata.xg[44]+lzs*(zdata.xg[45]+lzs*(zdata.xg[46]+lzs*(zdata.xg[47]+lzs*(zdata.xg[48]+lzs*zdata.xg[49]))))
    gbp[20] = zdata.xg[50]+lzs*(zdata.xg[51]+lzs*(zdata.xg[52]+lzs*(zdata.xg[53]+lzs*zdata.xg[54])))
    gbp[21] = zdata.xg[55]+lzs*(zdata.xg[56]+lzs*(zdata.xg[57]+lzs*(zdata.xg[58]+lzs*(zdata.xg[59]+lzs*zdata.xg[60]))))
    gbp[22] = zdata.xg[61]+lzs*(zdata.xg[62]+lzs*(zdata.xg[63]+lzs*(zdata.xg[64]+lzs*zdata.xg[65])))

    # Ragb
    gbp[23] = numpy.minimum(0.991640-743.123*z**2.83,
                  1.04220+lzs*(0.13156+0.045*lzs))
    gbp[24] = zdata.xg[66]+lzs*(zdata.xg[67]+lzs*(zdata.xg[68]+lzs*(zdata.xg[69]+ lzs*(zdata.xg[70]+lzs*zdata.xg[71]))))
    gbp[25] = zdata.xg[72]+lzs*(zdata.xg[73]+lzs*(zdata.xg[74]+lzs*(zdata.xg[75]+lzs*zdata.xg[76])))
    gbp[26] = zdata.xg[77]+lzs*(zdata.xg[78]+lzs*(zdata.xg[79]+lzs*(zdata.xg[80]+lzs*(zdata.xg[81]+lzs*zdata.xg[82]))))
    gbp[27] = zdata.xg[83]+lzs*(zdata.xg[84]+lzs*(zdata.xg[85]+lzs*(zdata.xg[86]+lzs*zdata.xg[87])))
    gbp[28] = zdata.xg[88]+lzs*(zdata.xg[89]+lzs*(zdata.xg[90]+lzs*(zdata.xg[91]+lzs*(zdata.xg[92]+lzs*zdata.xg[93]))))
    gbp[29] = zdata.xg[94]+lzs*(zdata.xg[95]+lzs*(zdata.xg[96]+lzs*(zdata.xg[97]+lzs*(zdata.xg[98]+lzs*zdata.xg[99]))))
    m1 = zpars[1] - 0.20
    gbp[30] = gbp[28] + gbp[29]*m1
    gbp[31] = numpy.minimum(gbp[24]/zpars[1]**gbp[25],gbp[26]/zpars[1]**gbp[27])

    # Mchei
    gbp[32] = zdata.xg[100]**4
    gbp[33] = zdata.xg[101]*4.0

    # Mcagb
    gbp[34] = zdata.xg[102]+lzs*(zdata.xg[103]+lzs*(zdata.xg[104]+lzs*zdata.xg[105]))
    gbp[35] = zdata.xg[106]+lzs*(zdata.xg[107]+lzs*(zdata.xg[108]+lzs*zdata.xg[109]))
    gbp[36] = zdata.xg[110]+lzs*zdata.xg[111]
    gbp[34] = gbp[34]**4
    gbp[35] = gbp[35]*4.0
    gbp[36] = gbp[36]**4

    # Lhei
    # set gbp[40] = -1.0 until it is reset later with an initial
    # call to Lheif using mass = zpars[1] and mhefl = 0.0
    gbp[37] = zdata.xh[0]+lzs*zdata.xh[1]
    gbp[38] = zdata.xh[2]+lzs*zdata.xh[3]
    gbp[39] = zdata.xh[4]
    gbp[40] = -1.0
    gbp[41] = zdata.xh[5]+lzs*(zdata.xh[6]+lzs*zdata.xh[7])
    gbp[42] = zdata.xh[8]+lzs*(zdata.xh[9]+lzs*zdata.xh[10])
    gbp[43] = zdata.xh[11]+lzs*(zdata.xh[12]+lzs*zdata.xh[13])
    gbp[41] = gbp[41]**2
    gbp[43] = gbp[43]**2
    # Lhe
    gbp[44] = zdata.xh[14]+lzs*(zdata.xh[15]+lzs*zdata.xh[16])
    if (lzs > -1.0):
        gbp[45] = 1.0 - zdata.xh[18]*(lzs+1.0)**zdata.xh[17]
    else:
        gbp[45] = 1.0

    gbp[46] = zdata.xh[19]+lzs*(zdata.xh[20]+lzs*zdata.xh[21])
    gbp[47] = zdata.xh[22]+lzs*(zdata.xh[23]+lzs*zdata.xh[24])
    gbp[44] = gbp[44]**gbp[47]
    gbp[46] = gbp[46]**gbp[47]
    gbp[45] = gbp[45]/zpars[2]**0.10+(gbp[45]*gbp[46]-gbp[44])/zpars[2]**(gbp[47]+0.10)

    # Rnumpy.minimum

    gbp[48] = zdata.xh[25]+lzs*(zdata.xh[26]+lzs*(zdata.xh[27]+lzs*zdata.xh[28]))
    gbp[49] = zdata.xh[29]+lzs*(zdata.xh[30]+lzs*(zdata.xh[31]+lzs*zdata.xh[32]))
    gbp[50] = zdata.xh[33]+lzs*(zdata.xh[34]+lzs*(zdata.xh[35]+lzs*zdata.xh[36]))
    gbp[51] = 5.0+zdata.xh[37]*z**zdata.xh[38]
    gbp[52] = zdata.xh[39]+lzs*(zdata.xh[40]+lzs*(zdata.xh[41]+lzs*zdata.xh[42]))
    gbp[48] = gbp[48]**gbp[52]
    gbp[50] = gbp[50]**(2.0*gbp[52])

    # The
    # set gbp[56] = -1.0 until it is reset later with an initial
    # call to Thef using mass = zpars[1], mc = 0.0  and mhefl = 0.0
    gbp[53] = zdata.xh[43]+lzs*(zdata.xh[44]+lzs*(zdata.xh[45]+lzs*zdata.xh[46]))
    gbp[54] = zdata.xh[47]+lzs*(zdata.xh[48]+lzs*zdata.xh[49])
    gbp[54] = numpy.maximum(gbp[54],1.0)
    gbp[55] = zdata.xh[50]
    gbp[56] = -1.0
    gbp[57] = zdata.xh[51]+lzs*(zdata.xh[52]+lzs*(zdata.xh[53]+lzs*zdata.xh[54]))
    gbp[58] = zdata.xh[55]+lzs*(zdata.xh[56]+lzs*(zdata.xh[57]+lzs*zdata.xh[58]))
    gbp[59] = zdata.xh[59]+lzs*(zdata.xh[60]+lzs*(zdata.xh[61]+lzs*zdata.xh[62]))
    gbp[60] = zdata.xh[63]+lzs*zdata.xh[64]
    gbp[57] = gbp[57]**gbp[60]
    gbp[59] = gbp[59]**5

    # Tbl
    dum1 = zpars[1]/zpars[2]
    gbp[61] = zdata.xh[65]+lzs*zdata.xh[66]
    gbp[61] = -gbp[61]*numpy.log10(dum1)
    gbp[62] = zdata.xh[67]
    if (lzd > 0.0):
        gbp[63] = 1.0-lzd*(zdata.xh[68]+lzd*(zdata.xh[69]+lzd*zdata.xh[70]))
    else:
        gbp[63] = 1.0

    gbp[64] = 1.0-gbp[63]*dum1**gbp[62]
    gbp[65] = 1.0 - lzd*(zdata.xh[76] + lzd*(zdata.xh[77] + lzd*zdata.xh[78]))
    gbp[66] = zdata.xh[71] + lzs*(zdata.xh[72] + lzs*(zdata.xh[73] + lzs*zdata.xh[74]))
    gbp[67] = zdata.xh[75]

    # Lzahb
    gbp[68] = zdata.xh[79] + lzs*(zdata.xh[80] + lzs*zdata.xh[81])
    gbp[69] = zdata.xh[82] + lzs*(zdata.xh[83] + lzs*zdata.xh[84])
    gbp[70] = 15.0
    gbp[71] = zdata.xh[85]
    gbp[72] = zdata.xh[86]

    # Rzahb
    gbp[74] = zdata.xh[87] + lzs*(zdata.xh[88] + lzs*(zdata.xh[89] + lzs*zdata.xh[90]))
    gbp[75] = zdata.xh[91] + lzs*(zdata.xh[92] + lzs*(zdata.xh[93] + lzs*zdata.xh[94]))
    gbp[76] = zdata.xh[95] + lzs*(zdata.xh[96] + lzs*(zdata.xh[97] + lzs*zdata.xh[98]))

    # finish Lbagb
    mhefl = 0.0
    lx = lbagbf(zpars[1], mhefl)
    gbp[15] = lx

    # finish LHeI
    dum1 = 0.0
    lhefl = lheif(zpars[1],mhefl)
    gbp[40] = (gbp[37]*zpars[1]**gbp[38]-lhefl)/(numpy.exp(zpars[1]*gbp[39])*lhefl)

    # finish THe
    thefl = thef(zpars[1],dum1,mhefl)*tbgbf(zpars[1])
    gbp[56] = (thefl-gbp[53])/(gbp[53]*numpy.exp(gbp[55]*zpars[1]))

    # finish Tblf
    rb = ragbf(zpars[2],lheif(zpars[2],zpars[1]),mhefl)
    rr = 1.0 - rnumpy.minimumf(zpars[2])/rb
    rr = numpy.maximum(rr,1.0e-12)
    gbp[65] = gbp[65]/(zpars[2]**gbp[66]*rr**gbp[67])

    # finish Lzahb
    gbp[73] = lhefl*lHef(zpars[1])

    kw = 0
    tm = 0.0
    tn = 0.0
    star(kw,zpars[1],zpars[1],tm,tn,tscls,lums,GB,zpars)
    zpars[8] = mcgbf(lums[2],GB,lums[5])
    zpars[9] = mcgbf(lums[3],GB,lums[5])

    # set the hydrogen and helium abundances
    zpars[10] = 0.760 - 3.0*z
    zpars[11] = 0.240 + 2.0*z

    # set constant for low-mass CHeB stars
    zpars[12] = (rnumpy.minimumf(zpars[1])/
                rgbf(zpars[1],lzahbf(zpars[1],zpars[8],zpars[1])))

    zpars[13] = z**0.40
    #
    """
    return zpars, msp



# -*- coding: utf-8 -*-
# Copyright (C) Katelyn Breivik (2017 - 2019)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`multidim`
"""

import numpy as np
import multiprocessing as mp
import math
import random
import scipy.integrate

from cosmic.utils import mass_min_max_select

from .sampler import register_sampler
from .. import InitialBinaryTable

from cosmic.utils import idl_tabulate, rndm

__author__ = 'Katelyn Breivik <katie.breivik@gmail.com>'
__credits__ = 'Scott Coughlin <scott.coughlin@ligo.org>'
__all__ = ['get_multidim_sampler','MultiDim']


def get_multidim_sampler(final_kstar1, final_kstar2, rand_seed, nproc, SFH_model, component_age, met, size, **kwargs):
    """adapted version of Maxwell Moe's IDL code that generates a population of single and binary stars

    Below is the adapted version of Maxwell Moe's IDL code
    that generates a population of single and binary stars
    based on the paper Mind your P's and Q's
    By Maxwell Moe and Rosanne Di Stefano

    The python code has been adopted by Mads Sørensen

    Version history:
    V. 0.1; 2017/02/03
    By Mads Sørensen
    - This is a pure adaption from IDL to Python.
    - The function idl_tabulate is similar to
    the IDL function int_tabulated except, this function seems to be slightly
    more exact in its solution.
    Therefore, relative to the IDL code, there are small numerical differences.

    Comments below beginning with ; is the original nodes by Maxwell Moe.
    Please read these careful for understanding the script.
    ; NOTE - This version produces only the statistical distributions of
    ;        single stars, binaries, and inner binaries in hierarchical triples.
    ;        Outer tertiaries in hierarchical triples are NOT generated.
    ;        Moreover, given a set of companions, all with period P to
    ;        primary mass M1, this version currently uses an approximation to
    ;        determine the fraction of those companions that are inner binaries
    ;        vs. outer triples. Nevertheless, this approximation reproduces
    ;        the overall multiplicity statistics.
    ; Step 1 - Tabulate probably density functions of periods,
    ;          mass ratios, and eccentricities based on
    ;          analytic fits to corrected binary star populations.
    ; Step 2 - Implement Monte Carlo method to generate stellar
    ;          population from those density functions.

    Parameters
    ----------
    final_kstar1 : `list` or `int`
        Int or list of final kstar1

    final_kstar2 : `list` or `int`
        Int or list of final kstar2

    rand_seed : `int`
        Int to seed random number generator

    nproc : `int`
        Number of processors to use to generate population

    SFH_model : `str`
        Model to sample star formation history (or birth time); choices include: const, burst, delta_burst

    component_age : `float`
        Sets the maximum age of the component; in the case of a delta burst, every binary is evolved for the component age

    met : `float`
        Sets the metallicity of the binary population where solar metallicity is 0.02

    size : `int`
        Size of the population to sample

    **porb_lo : `float`
        Lower limit in days for the orbital period distribution

    **porb_hi: `float`
        Upper limit in days for the orbital period distribution

    Returns
    -------
    InitialBinaryTable : `pandas.DataFrame`
        DataFrame in the format of the InitialBinaryTable

    mass_singles : `float`
        Total mass in single stars needed to generate population

    mass_binaries : `float`
        Total mass in binaries needed to generate population

    n_singles : `int`
        Number of single stars needed to generate a population

    n_binaries : `int`
        Number of binaries needed to generate a population
    """

    if type(final_kstar1) in [int, float]:
        final_kstar1 = [final_kstar1]
    if type(final_kstar2) in [int, float]:
        final_kstar2 = [final_kstar2]
    porb_lo = kwargs.pop('porb_lo', 0.15)
    porb_hi = kwargs.pop('porb_hi', 8.0)
    primary_min, primary_max, secondary_min, secondary_max = mass_min_max_select(final_kstar1, final_kstar2)
    initconditions = MultiDim()
    mass1_binary, mass2_binary, porb, ecc, mass_singles, mass_binaries, n_singles, n_binaries, binfrac = initconditions.initial_sample(primary_min, secondary_min, primary_max, secondary_max, porb_lo, porb_hi, rand_seed, size=size, nproc = nproc)
    tphysf, metallicity = initconditions.sample_SFH(SFH_model, component_age, met, size = mass1_binary.size)
    kstar1 = initconditions.set_kstar(mass1_binary)
    kstar2 = initconditions.set_kstar(mass2_binary)
    metallicity[metallicity < 1e-4] = 1e-4
    metallicity[metallicity > 0.03] = 0.03
    return InitialBinaryTable.InitialBinaries(mass1_binary, mass2_binary, porb, ecc, tphysf, kstar1, kstar2, metallicity, binfrac=binfrac), mass_singles, mass_binaries, n_singles, n_binaries

register_sampler('multidim', InitialBinaryTable, get_multidim_sampler,
                 usage="final_kstar1, final_kstar2, rand_seed, nproc, SFH_model, component_age, metallicity, size, binfrac")



class MultiDim:

    #-----------------------------------
    # Below is the adapted version of Maxwell Moe's IDL code
    # that generates a population of single and binary stars
    # based on the paper Mind your P's and Q's
    # By Maxwell Moe and Rosanne Di Stefano
    #
    # The python code has been adopted by Mads Sørensen
    #-----------------------------------
    # Version history:
    # V. 0.1; 2017/02/03
    # By Mads Sørensen
    # - This is a pure adaption from IDL to Python.
    # - The function idl_tabulate is similar to
    # the IDL function int_tabulated except, this function seems to be slightly
    # more exact in its solution.
    # Therefore, relative to the IDL code, there are small numerical differences.
    #-----------------------------------

    #
    # Comments below beginning with ; is the original nodes by Maxwell Moe.
    # Please read these careful for understanding the script.
    #; NOTE - This version produces only the statistical distributions of
    #;        single stars, binaries, and inner binaries in hierarchical triples.
    #;        Outer tertiaries in hierarchical triples are NOT generated.
    #;        Moreover, given a set of companions, all with period P to
    #;        primary mass M1, this version currently uses an approximation to
    #;        determine the fraction of those companions that are inner binaries
    #;        vs. outer triples. Nevertheless, this approximation reproduces
    #;        the overall multiplicity statistics.
    #; Step 1 - Tabulate probably density functions of periods,
    #;          mass ratios, and eccentricities based on
    #;          analytic fits to corrected binary star populations.
    #; Step 2 - Implement Monte Carlo method to generate stellar
    #;          population from those density functions.
    #;
    #
    #


    def initial_sample(self, M1min=0.08, M2min = 0.08, M1max=150.0, M2max=150.0, porb_lo=0.15, porb_hi=8.0, rand_seed=0, size=None, nproc=1):
        """Sample initial binary distribution according to Moe & Di Stefano (2017)
        <http://adsabs.harvard.edu/abs/2017ApJS..230...15M>`_

        Parameters
        ----------
        M1min : `float`
            minimum primary mass to sample [Msun]
            DEFAULT: 0.08
        M2min : `float`
            minimum secondary mass to sample [Msun]
            DEFAULT: 0.08
        M1max : `float`
            maximum primary mass to sample [Msun]
            DEFAULT: 150.0
        M2max : `float`
            maximum primary mass to sample [Msun]
            DEFAULT: 150.0
        porb_lo : `float`
            minimum orbital period to sample [log10(days)]
        porb_hi : `float`
            maximum orbital period to sample [log10(days)]
        rand_seed : int
            random seed generator
            DEFAULT: 0
        size : int, optional
            number of evolution times to sample
            NOTE: this is set in cosmic-pop call as Nstep

        Returns
        -------
        primary_mass_list : array
            array of primary masses with size=size
        secondary_mass_list : array
            array of secondary masses with size=size
        porb_list : array
            array of orbital periods in days with size=size
        ecc_list : array
            array of eccentricities with size=size
        mass_singles : `float`
            Total mass in single stars needed to generate population
        mass_binaries : `float`
            Total mass in binaries needed to generate population
        n_singles : `int`
            Number of single stars needed to generate a population
        n_binaries : `int`
            Number of binaries needed to generate a population
        binfrac_list : array
            array of binary probabilities based on primary mass and period with size=size
        """
        #Tabulate probably density functions of periods,
        #mass ratios, and eccentricities based on
        #analytic fits to corrected binary star populations.

        numM1 = 101
        #use binwidths to maintain structure of original array
        # default size is: numlogP=158
        bwlogP = 0.05
        numq = 91
        nume = 100

        #; Vector of primary masses M1 (Msun), logarithmic orbital period P (days),
        #; mass ratios q = Mcomp/M1, and eccentricities e
        #
        #; 0.8 < M1 < 40 (where we have statistics corrected for selection effects)
        M1_lo = 0.8
        M1_hi = 40

        M1v = np.logspace(np.log10(M1_lo), np.log10(M1_hi), numM1)
        #; 0.15 < log P < 8.0
        #; or use user specified values
        log10_porb_lo = porb_lo
        log10_porb_hi = porb_hi
        logPv = np.arange(log10_porb_lo, log10_porb_hi + bwlogP, bwlogP)
        numlogP = len(logPv)

        #; 0.10 < q < 1.00
        q_lo = 0.1
        q_hi = 1.0
        qv = np.linspace(q_lo, q_hi, numq)

        #; 0.0001 < e < 0.9901
        #; set minimum to non-zero value to avoid numerical errors
        e_lo = 0.0
        e_hi = 0.99
        ev = np.linspace(e_lo, e_hi, nume)+0.0001
        #; Note that companions outside this parameter space (e.g., q < 0.1,
        #; log P (days) > 8.0 are not constrained in M+D16 and therefore
        #; not considered.

        #; Distribution functions - define here, but evaluate within for loops.

        #; Frequency of companions with q > 0.1 per decade of orbital period.
        #; Bottom panel in Fig. 37 of M+D17
        flogP_sq = np.zeros([numlogP, numM1])

        #; Given M1 and P, the cumulative distribution of mass ratios q
        cumqdist = np.zeros([numq, numlogP, numM1])

        #; Given M1 and P, the cumulative distribution of eccentricities e
        cumedist = np.zeros([nume, numlogP, numM1])

        #; Given M1 and P, the probability that the companion
        #; is a member of the inner binary (currently an approximation).
        #; 100% for log P < 1.5, decreases with increasing P
        probbin = np.zeros([numlogP, numM1])


        #; Given M1, the cumulative period distribution of the inner binary
        #; Normalized so that max(cumPbindist) = total binary frac. (NOT unity)
        cumPbindist = np.zeros([numlogP, numM1])
        #; Slope alpha of period distribution across intermediate periods
        #; 2.7 - DlogP < log P < 2.7 + DlogP, see Section 9.3 and Eqn. 23.
        #; Slightly updated from version 1.
        alpha = 0.018
        DlogP = 0.7

        #; Heaviside function for twins with 0.95 < q < 1.00
        H = np.zeros(numq)
        ind = np.where(qv >= 0.95)
        H[ind] = 1.0
        H= H / idl_tabulate(qv, H) #;normalize so that integral is unity


        #; Relevant indices with respect to mass ratio
        indlq = np.where(qv >= 0.3)
        indsq = np.where(qv < 0.3)
        indq0p3 = np.min(indlq)

        # FILL IN THE MULTIDIMENSIONAL DISTRIBUTION FUNCTIONS
        #; Loop through primary mass
        for i in range(0, numM1):
            myM1 = M1v[i]
            #; Twin fraction parameters that are dependent on M1 only; section 9.1
            FtwinlogPle1 = 0.3 - 0.15 * np.log10(myM1)#; Eqn. 6
            logPtwin = 8.0 - myM1                       #; Eqn. 7a
            if (myM1 >= 6.5):
                logPtwin = 1.5                       #; Eqn. 7b
            #; Frequency of companions with q > 0.3 at different orbital periods
            #; and dependent on M1 only; section 9.3 (slightly modified since v1)
            flogPle1   = 0.020 + 0.04 * np.log10(myM1) + \
                         0.07 * (np.log10(myM1))**2.   #; Eqn. 20
            flogPeq2p7 = 0.039 + 0.07 * np.log10(myM1) + \
                         0.01 * (np.log10(myM1))**2.   #; Eqn. 21
            flogPeq5p5 = 0.078 - 0.05 * np.log10(myM1) + \
                         0.04 * (np.log10(myM1))**2.   #; Eqn. 22
            #; Loop through orbital period P
            for j in range(0, numlogP):
                mylogP = logPv[j]
                #; Given M1 and P, set excess twin fraction; section 9.1 and Eqn. 5
                if(mylogP <= 1.0):
                    Ftwin = FtwinlogPle1
                else:
                    Ftwin = FtwinlogPle1 * (1.0 - (mylogP - 1.0) / (logPtwin - 1.0))
                if(mylogP >= logPtwin):
                    Ftwin = 0.0


                #; Power-law slope gamma_largeq for M1 < 1.2 Msun and various P; Eqn. 9
                if(mylogP <= 5.0):
                    gl_1p2 = -0.5
                if(mylogP > 5.0):
                    gl_1p2 = -0.5 - 0.3 * (mylogP - 5.0)

                #; Power-law slope gamma_largeq for M1 = 3.5 Msun and various P; Eqn. 10
                if(mylogP <= 1.0):
                    gl_3p5 = -0.5
                if((mylogP > 1.0)and(mylogP <= 4.5)):
                    gl_3p5 = -0.5 - 0.2 * (mylogP - 1.0)
                if((mylogP > 4.5)and(mylogP <= 6.5)):
                    gl_3p5 = -1.2 - 0.4 * (mylogP - 4.5)
                if(mylogP > 6.5):
                    gl_3p5 = -2.0

                #; Power-law slope gamma_largeq for M1 > 6 Msun and various P; Eqn. 11
                if(mylogP <= 1.0):
                    gl_6 = -0.5
                if((mylogP > 1.0)and(mylogP <= 2.0)):
                    gl_6 = -0.5 - 0.9 * (mylogP - 1.0)
                if((mylogP > 2.0)and(mylogP <= 4.0)):
                    gl_6 = -1.4 - 0.3 * (mylogP - 2.0)

                if(mylogP > 4.0):
                    gl_6 = -2.0

                #; Given P, interpolate gamma_largeq w/ respect to M1 at myM1
                if(myM1 <= 1.2):
                    gl = gl_1p2
                if((myM1 > 1.2)and(myM1 <= 3.5)):
                    gl = np.interp(np.log10(myM1), np.log10([1.2, 3.5]), [gl_1p2, gl_3p5])
                if((myM1 > 3.5)and(myM1 <= 6.0)):
                    gl = np.interp(np.log10(myM1), np.log10([3.5, 6.0]), [gl_3p5, gl_6])
                if(myM1 > 6.0):
                    gl = gl_6

                #; Power-law slope gamma_smallq for M1 < 1.2 Msun and all P; Eqn. 13
                gs_1p2 = 0.3

                #; Power-law slope gamma_smallq for M1 = 3.5 Msun and various P; Eqn. 14
                if(mylogP <= 2.5):
                    gs_3p5 = 0.2
                if((mylogP > 2.5)and(mylogP <= 5.5)):
                    gs_3p5 = 0.2 - 0.3 * (mylogP - 2.5)
                if(mylogP > 5.5):
                    gs_3p5 = -0.7 - 0.2 * (mylogP - 5.5)

                #; Power-law slope gamma_smallq for M1 > 6 Msun and various P; Eqn. 15
                if(mylogP <= 1.0):
                    gs_6 = 0.1
                if((mylogP > 1.0)and(mylogP <= 3.0)):
                    gs_6 = 0.1 - 0.15 * (mylogP - 1.0)
                if((mylogP > 3.0)and(mylogP <= 5.6)):
                    gs_6 = -0.2 - 0.50 * (mylogP - 3.0)
                if(mylogP > 5.6):
                    gs_6 = -1.5

                #; Given P, interpolate gamma_smallq w/ respect to M1 at myM1
                if(myM1 <= 1.2):
                    gs = gs_1p2
                if((myM1 > 1.2)and(myM1 <= 3.5)):
                    gs = np.interp(np.log10(myM1), np.log10([1.2, 3.5]),[gs_1p2, gs_3p5])
                if((myM1 > 3.5)and(myM1 <= 6.0)):
                   gs = np.interp(np.log10(myM1), np.log10([3.5, 6.0]),[gs_3p5, gs_6])
                if(myM1 > 6.0):
                    gs = gs_6

                #; Given Ftwin, gamma_smallq, and gamma_largeq at the specified M1 & P,
                #; tabulate the cumulative mass ratio distribution across 0.1 < q < 1.0
                fq = qv**gl                                   #; slope across 0.3 < q < 1.0
                fq = fq / idl_tabulate(qv[indlq], fq[indlq])   #; normalize to 0.3 < q < 1.0
                fq = fq * (1.0 - Ftwin) + H * Ftwin                   #; add twins
                fq[indsq] = fq[indq0p3] * (qv[indsq] / 0.3)**gs   #; slope across 0.1 < q < 0.3
                cumfq = np.cumsum(fq) - fq[0]          #; cumulative distribution
                cumfq = cumfq / np.max(cumfq)                     #; normalize cumfq(q=1.0) = 1
                cumqdist[:,j,i] = cumfq                      #; save to grid

                #; Given M1 and P, q_factor is the ratio of all binaries 0.1 < q < 1.0
                #; to those with 0.3 < q < 1.0
                q_factor = idl_tabulate(qv, fq)


                #; Given M1 & P, calculate power-law slope eta of eccentricity dist.
                if(mylogP >= 0.7):
                    #; For log P > 0.7 use fits in Section 9.2.
                    #; Power-law slope eta for M1 < 3 Msun and log P > 0.7
                    eta_3 = 0.6 - 0.7 / (mylogP - 0.5)  #; Eqn. 17
                    #; Power-law slope eta for M1 > 7 Msun and log P > 0.7
                    eta_7 = 0.9 - 0.2 / (mylogP - 0.5)  #; Eqn. 18
                else:
                    #; For log P < 0.7, set eta to fitted values at log P = 0.7
                    eta_3 = -2.9
                    eta_7 = -0.1

                #; Given P, interpolate eta with respect to M1 at myM1
                if(myM1 <= 3.):
                    eta = eta_3
                if((myM1 > 3.)and(myM1 <= 7.)):
                    eta = np.interp(np.log10(myM1), np.log10([3., 7.]), [eta_3, eta_7])
                if(myM1 > 7.):
                    eta = eta_7


                #; Given eta at the specified M1 and P, tabulate eccentricity distribution
                if(10**mylogP <= 2.):
                    #; For P < 2 days, assume all systems are close to circular
                    #; For adopted ev (spacing and minimum value), eta = -3.2 satisfies this
                    fe = ev**(-3.2)
                else:
                    fe = ev**eta
                    e_max = 1.0 - (10**mylogP / 2.0)**(-2.0/3.0) #; maximum eccentricity for given P
                    ind = np.where(ev >= e_max)
                    fe[ind] = 0.0                       #; set dist. = 0 for e > e_max
                    #; Assume e dist. has power-law slope eta for 0.0 < e / e_max < 0.8 and
                    #; then linear turnover between 0.8 < e / e_max < 1.0 so that dist.
                    #; is continuous at e / e_max = 0.8 and zero at e = e_max
                    ind = np.where((ev >= 0.8*e_max)&(ev <= 1.0*e_max))
                    ind_cont = np.min(ind) - 1
                    fe[ind] = np.interp(ev[ind], [0.8*e_max, 1.0*e_max], [fe[ind_cont], 0.])

                cumfe = np.cumsum(fe) - fe[0]  #; cumulative distribution
                cumfe = cumfe / np.max(cumfe)             #; normalize cumfe(e=e_max) = 1
                cumedist[:,j,i] = cumfe              #; save to grid


                #; Given constants alpha and DlogP and
                #; M1 dependent values flogPle1, flogPeq2p7, and flogPeq5p5,
                #; calculate frequency flogP of companions with q > 0.3 per decade
                #; of orbital period at given P (Section 9.3 and Eqn. 23)
                if(mylogP <= 1.):
                    flogP = flogPle1
                if((mylogP > 1.0)and(mylogP <= 2.7 - DlogP)):
                    flogP = flogPle1 + (mylogP - 1.0) / (1.7 - DlogP) * \
                            (flogPeq2p7 - flogPle1 - alpha*DlogP)
                if((mylogP > 2.7 - DlogP)and(mylogP <= 2.7 + DlogP)):
                    flogP = flogPeq2p7 + alpha*(mylogP - 2.7)
                if((mylogP > 2.7 + DlogP)and(mylogP <= 5.5)):
                    flogP = flogPeq2p7 + alpha*DlogP + \
                            (mylogP - 2.7 - DlogP)/(2.8 - DlogP) * \
                            (flogPeq5p5 - flogPeq2p7 - alpha*DlogP)
                if(mylogP > 5.5):
                    flogP = flogPeq5p5 * np.exp(-0.3 * (mylogP - 5.5))


                #; Convert frequency of companions with q > 0.3 to frequency of
                #; companions with q > 0.1 according to q_factor; save to grid
                flogP_sq[j,i] = flogP*q_factor


                #; Calculate prob. that a companion to M1 with period P is the
                #; inner binary.  Currently this is an approximation.
                #; 100% for log P < 1.5
                #; For log P > 1.5 adopt functional form that reproduces M1 dependent
                #; multiplicity statistics in Section 9.4, including a
                #; 41% binary star faction (59% single star fraction) for M1 = 1 Msun and
                #; 96% binary star fraction (4% single star fraction) for M1 = 28 Msun
                if(mylogP <= 1.5):
                    probbin[j,i] = 1.0
                else:
                    probbin[j,i] = 1.0 - 0.11*(mylogP - 1.5)**1.43 * (myM1 / 10.0)**0.56
                if(probbin[j,i] <= 0.0):
                    probbin[j,i]=0.0

            #; Given M1, calculate cumulative binary period distribution
            mycumPbindist = np.cumsum(flogP_sq[:,i] * probbin[:,i]) - \
                            flogP_sq[0,i] * probbin[0,i]
            #; Normalize so that max(cumPbindist) = total binary star fraction (NOT 1)
            mycumPbindist = mycumPbindist / np.max(mycumPbindist) * \
                            idl_tabulate(logPv, flogP_sq[:,i]*probbin[:,i])
            cumPbindist[:,i] = mycumPbindist  #;save to grid

        #; Step 2
        #; Implement Monte Carlo method / random number generator to select
        #; single stars and binaries from the grids of distributions


        #; Create vector for PRIMARY mass function, which is the mass distribution
        #; of single stars and primaries in binaries.
        #; This is NOT the IMF, which is the mass distribution of single stars,
        #; primaries in binaries, and secondaries in binaries.

        primary_mass_list = []
        secondary_mass_list = []
        porb_list = []
        ecc_list = []


        def _sample_initial_pop(M1min, M2min, M1max, M2max, size, nproc, seed, output):
            # get unique and replicatable seed for each process
            process = mp.Process()
            mp_seed = (process._identity[0]-1)+(nproc*(process._identity[1]-1))
            np.random.seed(seed + mp_seed)

            mass_singles = 0.0
            mass_binaries = 0.0
            n_singles = 0
            n_binaries = 0
            primary_mass_list = []
            secondary_mass_list = []
            porb_list = []
            ecc_list = []
            binfrac_list = []

            #; Full primary mass vector across 0.08 < M1 < 150
            M1 = np.linspace(0,150,150000) + 0.08
            #; Slope = -2.3 for M1 > 1 Msun
            fM1 = M1**(-2.3)
            #; Slope = -1.6 for M1 = 0.5 - 1.0 Msun
            ind = np.where(M1 <= 1.0)
            fM1[ind] = M1[ind]**(-1.6)
            #; Slope = -0.8 for M1 = 0.15 - 0.5 Msun
            ind = np.where(M1 <= 0.5)
            fM1[ind] = M1[ind]**(-0.8) / 0.5**(1.6 - 0.8)
            #; Cumulative primary mass distribution function
            cumfM1 = np.cumsum(fM1) - fM1[0]
            cumfM1 = cumfM1 / np.max(cumfM1)
            #; Value of primary mass CDF where M1 = M1min
            #; Minimum primary mass to generate (must be >0.080 Msun)
            cumf_M1min = np.interp(0.08, M1, cumfM1)
            while len(primary_mass_list) < size:

                #; Select primary M1 > M1min from primary mass function
                myM1 = np.interp(cumf_M1min + (1.0 - cumf_M1min) * np.random.rand(), cumfM1, M1)


                # ; Find index of M1v that is closest to myM1.
                #     ; For M1 = 40 - 150 Msun, adopt binary statistics of M1 = 40 Msun.
                #     ; For M1 = 0.08 - 0.8 Msun, adopt P and e dist of M1 = 0.8Msun,
                #     ; scale and interpolate the companion frequencies so that the
                #     ; binary star fraction of M1 = 0.08 Msun primaries is zero,
                #     ; and truncate the q distribution so that q > q_min = 0.08/M1
                indM1 = np.where(abs(myM1 - M1v) == min(abs(myM1 - M1v)))
                indM1 = indM1[0]


                # ; Given M1, determine cumulative binary period distribution
                mycumPbindist_flat = (cumPbindist[:, indM1]).flatten()
                #; If M1 < 0.8 Msun, rescale to appropriate binary star fraction
                if(myM1 <= 0.8):
                    mycumPbindist_flat = mycumPbindist_flat * np.interp(np.log10(myM1), np.log10([0.08, 0.8]), [0.0, 1.0])

                # ; Given M1, determine the binary star fraction
                mybinfrac = np.max(mycumPbindist_flat)


                # ; Generate random number myrand between 0 and 1
                myrand = np.random.rand()
                #; If random number < binary star fraction, generate a binary
                if(myrand < mybinfrac):
                    #; Given myrand, select P and corresponding index in logPv
                    mylogP = np.interp(myrand, mycumPbindist_flat, logPv)
                    indlogP = np.where(abs(mylogP - logPv) == min(abs(mylogP - logPv)))
                    indlogP = indlogP[0]


                    #; Given M1 & P, select e from eccentricity distribution
                    mye = np.interp(np.random.rand(), cumedist[:, indlogP, indM1].flatten(), ev)


                    #; Given M1 & P, determine mass ratio distribution.
                    #; If M1 < 0.8 Msun, truncate q distribution and consider
                    #; only mass ratios q > q_min = 0.08 / M1
                    mycumqdist = cumqdist[:, indlogP, indM1].flatten()
                    if(myM1 < 0.8):
                        q_min = 0.08 / myM1
                        #; Calculate cumulative probability at q = q_min
                        cum_qmin = np.interp(q_min, qv, mycumqdist)
                        #; Rescale and renormalize cumulative distribution for q > q_min
                        mycumqdist = mycumqdist - cum_qmin
                        mycumqdist = mycumqdist / max(mycumqdist)
                        #; Set probability = 0 where q < q_min
                        indq = np.where(qv <= q_min)
                        mycumqdist[indq] = 0.0

                    #; Given M1 & P, select q from cumulative mass ratio distribution
                    myq = np.interp(np.random.rand(), mycumqdist, qv)

                    if myM1 > M1min and myq * myM1 > M2min and myM1 < M1max and myq * myM1 < M2max and mylogP < porb_hi and mylogP > porb_lo:
                        primary_mass_list.append(myM1)
                        secondary_mass_list.append(myq * myM1)
                        porb_list.append(10**mylogP)
                        ecc_list.append(mye)
                        binfrac_list.append(mybinfrac)
                    mass_binaries += myM1
                    mass_binaries += myq * myM1
                    n_binaries += 1
                else:
                    mass_singles += myM1
                    n_singles += 1
            output.put([primary_mass_list, secondary_mass_list, porb_list, ecc_list, mass_singles, mass_binaries, n_singles, n_binaries, binfrac_list])
            return

        output = mp.Queue()
        processes = [mp.Process(target = _sample_initial_pop,\
                                args = (M1min, M2min, M1max, M2max, size/nproc, nproc, rand_seed, output))\
                                for x in range(nproc)]
        for p in processes:
            p.daemon = True
            p.start()
        results = [output.get() for p in processes]
        for p in processes:
            p.join()

        primary_mass_list = []
        secondary_mass_list = []
        porb_list = []
        ecc_list = []
        mass_singles = []
        mass_binaries = []
        n_singles = []
        n_binaries = []
        binfrac_list = []
        dat_lists = [[],[],[],[],[],[],[],[],[]]

        for output_list in results:
            ii = 0
            for dat_list in output_list:
               dat_lists[ii].append(dat_list)
               ii+=1

        primary_mass_list = np.hstack(dat_lists[0])
        secondary_mass_list = np.hstack(dat_lists[1])
        porb_list = np.hstack(dat_lists[2])
        ecc_list = np.hstack(dat_lists[3])
        mass_singles = np.sum(dat_lists[4])
        mass_binaries = np.sum(dat_lists[5])
        n_singles = np.sum(dat_lists[6])
        n_binaries = np.sum(dat_lists[7])
        binfrac_list = np.hstack(dat_lists[8])

        return primary_mass_list, secondary_mass_list, porb_list, ecc_list, mass_singles, mass_binaries, n_singles, n_binaries, binfrac_list

    def sample_SFH(self, SFH_model='const', component_age=10000.0, met = 0.02, size=None):
        """Sample an evolution time for each binary based on a user-specified
        star formation history (SFH) and Galactic component age.
        The default is a MW thin disk constant evolution over 10000 Myr

        Parameters
        ----------
        SFH_model : string
            'const' assigns an evolution time assuming a constant star
            formation rate over the age of the MW disk: component_age [Myr]
            'burst' assigns an evolution time assuming a burst of constant
            star formation for 1Gyr starting at component_age [Myr] in the past
            'delta_burst' assignes a t=0 evolution time until component age
            Default: 'const'
        component_age : float
            age of the Galactic component [Myr]
            Default: 10000.0
        met : float
            metallicity of the population [Z_sun = 0.02]
            Default: 0.02
        size : int, optional
            number of evolution times to sample
            NOTE: this is set in cosmic-pop call as Nstep

        Returns
        -------
        tphys : array
            array of evolution times of size=size
        """

        if SFH_model=='const':

            tphys = np.random.uniform(0, component_age, size)
            metallicity = np.ones(size)*met
            return tphys, metallicity

        elif SFH_model=='burst':
            tphys = component_age - np.random.uniform(0, 1000, size)
            metallicity = np.ones(size)*met
            return tphys, metallicity
        elif SFH_model=='delta_burst':
            tphys = component_age*np.ones(size)
            metallicity = np.ones(size)*met
            return tphys, metallicity

    def set_kstar(self, mass):
        """Initialize stellar types according to BSE classification
        kstar=1 if M>=0.7 Msun; kstar=0 if M<0.7 Msun

        Parameters
        ----------
        mass : array
            array of masses

        Returns
        -------
        kstar : array
            array of initial stellar types
        """

        kstar = np.zeros(mass.size)
        low_cutoff = 0.7
        lowIdx = np.where(mass < low_cutoff)[0]
        hiIdx = np.where(mass >= low_cutoff)[0]

        kstar[lowIdx] = 0
        kstar[hiIdx] = 1

        return kstar
